import os
from logging import getLogger

from PyQt5.QtCore import QObject, pyqtSignal
from legendary.lfs.eos import EOSOverlayApp
from legendary.models.downloading import ConditionCheckResult

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.core import LegendaryCore
from rare.lgndr.glue.arguments import LgndrInstallGameArgs
from rare.lgndr.glue.exception import LgndrException
from rare.lgndr.glue.monkeys import LgndrIndirectStatus
from rare.models.install import InstallDownloadModel, InstallOptionsModel
from .worker import Worker

logger = getLogger("InstallInfoWorker")


class InstallInfoWorker(Worker):
    class Signals(QObject):
        result = pyqtSignal(InstallDownloadModel)
        failed = pyqtSignal(str)
        finished = pyqtSignal()

    def __init__(self, core: LegendaryCore, options: InstallOptionsModel):
        super(InstallInfoWorker, self).__init__()
        self.signals = InstallInfoWorker.Signals()
        self.core = core
        self.options = options

    def run_real(self):
        try:
            if not self.options.overlay:
                cli = LegendaryCLI(self.core)
                status = LgndrIndirectStatus()
                result = cli.install_game(
                    LgndrInstallGameArgs(**self.options.as_install_kwargs(), indirect_status=status)
                )
                if result:
                    download = InstallDownloadModel(*result)
                else:
                    raise LgndrException(status.message)
            else:
                if not os.path.exists(path := self.options.base_path):
                    os.makedirs(path)

                dlm, analysis, igame = self.core.prepare_overlay_install(
                    path=self.options.base_path
                )

                download = InstallDownloadModel(
                    dlm=dlm,
                    analysis=analysis,
                    igame=igame,
                    game=EOSOverlayApp,
                    repair=False,
                    repair_file="",
                    res=ConditionCheckResult(),  # empty
                )

            if not download.res or not download.res.failures:
                self.signals.result.emit(download)
            else:
                self.signals.failed.emit("\n".join(str(i) for i in download.res.failures))
        except LgndrException as ret:
            self.signals.failed.emit(ret.message)
        except Exception as e:
            self.signals.failed.emit(str(e))
        self.signals.finished.emit()



from enum import Enum

from PyQt5.QtCore import QSize


class Orientation(Enum):
    Tall = 0
    Wide = 1


class ImageSize:
    class Preset:
        def __init__(self, divisor: float, pixel_ratio: float, orientation: Orientation = Orientation.Tall,
                     base: 'ImageSize.Preset' = None):
            self.__divisor = divisor
            self.__pixel_ratio = pixel_ratio
            if orientation == Orientation.Tall:
                self.__img_factor = 67
                self.__size = QSize(self.__img_factor * 3, self.__img_factor * 4) * pixel_ratio / divisor
            else:
                self.__img_factor = 17
                self.__size = QSize(self.__img_factor * 16, self.__img_factor * 9) * pixel_ratio / divisor
            # lk: for prettier images set this to true
            self.__smooth_transform: bool = True
            if divisor > 2:
                self.__smooth_transform = False

            if base is not None:
                self.__base = base
            else:
                self.__base = self

        def __eq__(self, other: 'ImageSize.Preset'):
            return (
                self.__size == other.size
                and self.__divisor == other.divisor
                and self.__smooth_transform == other.smooth
                and self.__pixel_ratio == other.pixel_ratio
            )

        @property
        def size(self) -> QSize:
            return self.__size

        @property
        def divisor(self) -> float:
            return self.__divisor

        @property
        def smooth(self) -> bool:
            return self.__smooth_transform

        @property
        def pixel_ratio(self) -> float:
            return self.__pixel_ratio

        @property
        def base(self) -> 'ImageSize.Preset':
            return self.__base

    Image = Preset(1, 2)
    """! @brief Size and pixel ratio of the image on disk"""

    ImageWide = Preset(1, 2, Orientation.Wide)
    """! @brief Size and pixel ratio for wide 16/9 image on disk"""

    Display = Preset(1, 1, base=Image)
    """! @brief Size and pixel ratio for displaying"""

    DisplayWide = Preset(1, 1, Orientation.Wide, base=ImageWide)
    """! @brief Size and pixel ratio for wide 16/9 image display"""

    LibraryWide = Preset(1.21, 1, Orientation.Wide, base=ImageWide)

    Library = Preset(1.21, 1, base=Image)
    """! @brief Same as Display"""

    Small = Preset(3, 1, base=Image)
    """! @brief Small image size for displaying"""

    SmallWide = Preset(3, 1, Orientation.Wide, base=ImageWide)
    """! @brief Small image size for displaying"""

    Smaller = Preset(4, 1, base=Image)
    """! @brief Smaller image size for displaying"""

    SmallerWide = Preset(4, 1, Orientation.Wide, base=ImageWide)
    """! @brief Smaller image size for displaying"""

    Icon = Preset(5, 1, base=Image)
    """! @brief Smaller image size for UI icons"""



import json
import logging
from enum import IntEnum

from PyQt5.QtCore import QObject, pyqtSignal, QTimer, pyqtSlot
from PyQt5.QtNetwork import QLocalSocket
from PyQt5.QtWidgets import QMessageBox
from legendary.models.game import Game

from rare.models.launcher import ErrorModel, Actions, FinishedModel, StateChangedModel

logger = logging.getLogger("GameProcess")


class GameProcess(QObject):
    # int: code
    launched = pyqtSignal(int)
    # int: code
    finished = pyqtSignal(int)

    class Code(IntEnum):
        SUCCESS = 0
        ERROR = 1
        TIMEOUT = -1
        ON_STARTUP = -1234

    def __init__(self, game: Game):
        super(GameProcess, self).__init__()
        self.game = game
        self.on_startup = False
        self.tried_connections = 0

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.__connect)
        self.socket = QLocalSocket()
        self.socket_name = f"rare_{self.game.app_name}"

        self.socket.connected.connect(self.__on_connected)
        try:
            self.socket.errorOccurred.connect(self.__on_error)
        except AttributeError:
            # QTimer.singleShot(
            #     100,
            #     lambda: self._error_occurred(QLocalSocket.UnknownSocketError) if self.socket.error() else None
            # )
            logger.warning("Do not handle errors on QLocalSocket, because of an old qt version")
        self.socket.readyRead.connect(self.__on_message)
        self.socket.disconnected.connect(self.__close)

    def connect_to_server(self, on_startup: bool):
        self.on_startup = on_startup
        self.timer.start(200)

    @pyqtSlot()
    def __close(self):
        try:
            self.socket.close()
        except RuntimeError:
            pass

    @pyqtSlot()
    def __connect(self):
        self.socket.connectToServer(self.socket_name)
        self.tried_connections += 1

        if self.tried_connections > 50:  # 10 seconds
            QMessageBox.warning(None, "Error", self.tr("Connection to game process failed (Timeout)"))
            self.timer.stop()
            self.finished.emit(GameProcess.Code.TIMEOUT)

    @pyqtSlot()
    def __on_message(self):
        message = self.socket.readAll().data()
        if not message.startswith(b"{"):
            logger.error(f"Received unsupported message: {message.decode('utf-8')}")
            return
        try:
            data = json.loads(message)
        except json.JSONDecodeError as e:
            logger.error(e)
            logger.error("Could not load json data")
            return

        action = data.get("action", False)

        if not action:
            logger.error("Got unexpected action")
        elif action == Actions.finished:
            logger.info(f"{self.game.app_name} {self.game.app_title} finished")
            model = FinishedModel.from_json(data)
            self.socket.close()
            self.__game_finished(model.exit_code)
        elif action == Actions.error:
            model = ErrorModel.from_json(data)
            logger.error(f"Error in game {self.game.app_title}: {model.error_string}")
            self.socket.close()
            self.__game_finished(GameProcess.Code.ERROR)
            QMessageBox.warning(None, "Error", self.tr(
                "Error in game {}:\n{}").format(self.game.app_title, model.error_string))
        elif action == Actions.state_update:
            model = StateChangedModel.from_json(data)
            if model.new_state == StateChangedModel.States.started:
                logger.info("Launched Game")
                self.launched.emit(GameProcess.Code.SUCCESS)

    @pyqtSlot()
    def __on_connected(self):
        self.timer.stop()
        logger.info(f"Connection established for {self.game.app_name} ({self.game.app_title})")
        if self.on_startup:
            logger.info(f"Found {self.game.app_name} ({self.game.app_title}) running at startup")
            self.launched.emit(GameProcess.Code.ON_STARTUP)

    @pyqtSlot(QLocalSocket.LocalSocketError)
    def __on_error(self, _: QLocalSocket.LocalSocketError):
        if self.on_startup:
            self.timer.stop()
            self.__close()
            self.__game_finished(GameProcess.Code.ON_STARTUP)  # 1234 is exit code for startup
        else:
            logger.error(f"{self.game.app_name} ({self.game.app_title}): {self.socket.errorString()}")

    def __game_finished(self, exit_code: int):
        self.finished.emit(exit_code)



import json
import os
import platform
from dataclasses import dataclass, field
from datetime import datetime
from logging import getLogger
from threading import Lock
from typing import List, Optional, Dict, Set

from PyQt5.QtCore import QRunnable, pyqtSlot, QProcess, QThreadPool
from PyQt5.QtGui import QPixmap
from legendary.models.game import Game, InstalledGame
from legendary.utils.selective_dl import get_sdl_appname

from rare.lgndr.core import LegendaryCore
from rare.models.install import InstallOptionsModel, UninstallOptionsModel
from rare.models.base_game import RareGameBase, RareGameSlim
from rare.shared.game_process import GameProcess
from rare.shared.image_manager import ImageManager
from rare.utils.paths import data_dir, get_rare_executable
from rare.utils.steam_grades import get_rating

logger = getLogger("RareGame")


class RareGame(RareGameSlim):
    @dataclass
    class Metadata:
        auto_update: bool = False
        queued: bool = False
        queue_pos: Optional[int] = None
        last_played: datetime = datetime.min
        grant_date: Optional[datetime] = None
        steam_grade: Optional[str] = None
        steam_date: datetime = datetime.min
        tags: List[str] = field(default_factory=list)

        @classmethod
        def from_dict(cls, data: Dict):
            return cls(
                auto_update=data.get("auto_update", False),
                queued=data.get("queued", False),
                queue_pos=data.get("queue_pos", None),
                last_played=datetime.fromisoformat(data["last_played"]) if data.get("last_played", None) else datetime.min,
                grant_date=datetime.fromisoformat(data["grant_date"]) if data.get("grant_date", None) else None,
                steam_grade=data.get("steam_grade", None),
                steam_date=datetime.fromisoformat(data["steam_date"]) if data.get("steam_date", None) else datetime.min,
                tags=data.get("tags", []),
            )

        def as_dict(self):
            return dict(
                auto_update=self.auto_update,
                queued=self.queued,
                queue_pos=self.queue_pos,
                last_played=self.last_played.isoformat() if self.last_played else datetime.min,
                grant_date=self.grant_date.isoformat() if self.grant_date else None,
                steam_grade=self.steam_grade,
                steam_date=self.steam_date.isoformat() if self.steam_date else datetime.min,
                tags=self.tags,
            )

        def __bool__(self):
            return self.queued or self.queue_pos is not None or self.last_played is not None

    def __init__(self, legendary_core: LegendaryCore, image_manager: ImageManager, game: Game):
        super(RareGame, self).__init__(legendary_core, game)
        self.__origin_install_path: Optional[str] = None
        self.__origin_install_size: Optional[int] = None

        self.image_manager = image_manager

        # Update names for Unreal Engine
        if self.game.app_title == "Unreal Engine":
            self.game.app_title += f" {self.game.app_name.split('_')[-1]}"

        self.pixmap: QPixmap = QPixmap()
        self.metadata: RareGame.Metadata = RareGame.Metadata()
        self.__load_metadata()

        self.owned_dlcs: Set[RareGame] = set()

        if self.has_update:
            logger.info(f"Update available for game: {self.app_name} ({self.app_title})")

        self.__worker: Optional[QRunnable] = None
        self.progress: int = 0
        self.signals.progress.start.connect(lambda: self.__on_progress_update(0))
        self.signals.progress.update.connect(self.__on_progress_update)

        self.game_process = GameProcess(self.game)
        self.game_process.launched.connect(self.__game_launched)
        self.game_process.finished.connect(self.__game_finished)
        if self.is_installed and not self.is_dlc:
            self.game_process.connect_to_server(on_startup=True)

    def add_dlc(self, dlc) -> None:
        # lk: plug dlc progress signals to the game's
        dlc.signals.progress.start.connect(self.signals.progress.start)
        dlc.signals.progress.update.connect(self.signals.progress.update)
        dlc.signals.progress.finish.connect(self.signals.progress.finish)
        self.owned_dlcs.add(dlc)

    def __on_progress_update(self, progress: int):
        self.progress = progress

    def worker(self) -> Optional[QRunnable]:
        return self.__worker

    def set_worker(self, worker: Optional[QRunnable]):
        self.__worker = worker
        if worker is None:
            self.state = RareGame.State.IDLE

    @pyqtSlot(int)
    def __game_launched(self, code: int):
        self.state = RareGame.State.RUNNING
        self.metadata.last_played = datetime.now()
        if code == GameProcess.Code.ON_STARTUP:
            return
        self.__save_metadata()
        self.signals.game.launched.emit(self.app_name)

    @pyqtSlot(int)
    def __game_finished(self, exit_code: int):
        if exit_code == GameProcess.Code.ON_STARTUP:
            return
        if self.supports_cloud_saves:
            self.update_saves()
        self.state = RareGame.State.IDLE
        self.signals.game.finished.emit(self.app_name)

    __metadata_json: Optional[Dict] = None
    __metadata_lock: Lock = Lock()

    @staticmethod
    def __load_metadata_json() -> Dict:
        if RareGame.__metadata_json is None:
            metadata = {}
            try:
                with open(os.path.join(data_dir(), "game_meta.json"), "r") as metadata_fh:
                    metadata = json.load(metadata_fh)
            except FileNotFoundError:
                logger.info("Game metadata json file does not exist.")
            except json.JSONDecodeError:
                logger.warning("Game metadata json file is corrupt.")
            finally:
                RareGame.__metadata_json = metadata
        return RareGame.__metadata_json

    def __load_metadata(self):
        with RareGame.__metadata_lock:
            metadata: Dict = self.__load_metadata_json()
            # pylint: disable=unsupported-membership-test
            if self.app_name in metadata:
                # pylint: disable=unsubscriptable-object
                self.metadata = RareGame.Metadata.from_dict(metadata[self.app_name])

    def __save_metadata(self):
        with RareGame.__metadata_lock:
            metadata: Dict = self.__load_metadata_json()
            # pylint: disable=unsupported-assignment-operation
            metadata[self.app_name] = self.metadata.as_dict()
            with open(os.path.join(data_dir(), "game_meta.json"), "w") as metadata_json:
                json.dump(metadata, metadata_json, indent=2)

    def update_game(self):
        self.game = self.core.get_game(
            self.app_name, update_meta=False,
            platform=self.igame.platform if self.igame else self.default_platform
        )

    def update_igame(self):
        self.igame = self.core.get_installed_game(self.app_name)

    def store_igame(self):
        self.core.lgd.set_installed_game(self.app_name, self.igame)
        self.update_igame()

    def update_rgame(self):
        self.update_game()
        self.update_igame()

    @property
    def developer(self) -> str:
        """!
        @brief Property to report the developer of a Game

        @return str
        """
        return self.game.metadata["developer"]

    @property
    def install_size(self) -> int:
        """!
        @brief Property to report the installation size of an InstalledGame

        @return int The size of the installation
        """
        if self.is_origin:
            return self.__origin_install_size if self.__origin_install_size is not None else 0
        return self.igame.install_size if self.igame is not None else 0

    @property
    def install_path(self) -> Optional[str]:
        if self.is_origin:
            # TODO Linux is also C:\\...
            return self.__origin_install_path
        return super(RareGame, self).install_path

    @install_path.setter
    def install_path(self, path: str) -> None:
        if self.igame:
            self.igame.install_path = path
            self.store_igame()
        elif self.is_origin:
            self.__origin_install_path = path

    @property
    def remote_version(self) -> str:
        """!
        @brief Property to report the remote version of an InstalledGame

        If the Game is installed, requests the latest version string from EGS,
        otherwise it reports the already known version of the Game for Windows.

        @return str The current version from EGS
        """
        if self.igame is not None:
            return self.game.app_version(self.igame.platform)
        else:
            return self.game.app_version(self.default_platform)

    @property
    def has_update(self) -> bool:
        """!
        @brief Property to report if an InstalledGame has updates available

        Games have to be installed and have assets available to have
        updates

        @return bool If there is an update available
        """
        if self.igame is not None and self.core.lgd.assets is not None:
            try:
                if self.remote_version != self.igame.version:
                    return True
            except ValueError:
                logger.error(f"Asset error for {self.game.app_title}")
                return False
        return False

    @property
    def is_installed(self) -> bool:
        """!
        @brief Property to report if a game is installed

        This returns True if InstalledGame data have been loaded for the game
        or if the game is a game without assets, for example an Origin game.

        @return bool If the game should be considered installed
        """
        return (self.igame is not None) \
            or (self.is_origin and self.__origin_install_path is not None)

    def set_installed(self, installed: bool) -> None:
        """!
        @brief Sets the installation status of a game

        If this is set to True the InstalledGame data is fetched
        for the game, if set to False the igame attribute is cleared.

        @param installed The installation status of the game
        @return None
        """
        if installed:
            self.update_igame()
            if not self.is_dlc:
                self.core.egstore_delete(self.igame)
                self.core.egstore_write(self.igame.app_name)
            self.signals.game.installed.emit(self.app_name)
            if self.has_update:
                self.signals.download.enqueue.emit(self.app_name)
        else:
            if self.has_update:
                self.signals.download.dequeue.emit(self.app_name)
            self.core.egstore_delete(self.igame)
            self.igame = None
            self.signals.game.uninstalled.emit(self.app_name)
        self.set_pixmap()

    @property
    def can_run_offline(self) -> bool:
        """!
        @brief Property to report if a game can run offline

        Checks if the game can run without connectin the internet.
        It's a simple wrapper around legendary provided information,
        with handling of not installed games.

        @return bool If the games can run without network
        """
        return self.igame.can_run_offline if self.igame is not None else False

    @property
    def is_foreign(self) -> bool:
        """!
        @brief Property to report if a game doesn't belong to the current account

        Checks if a game belongs to the currently logged in account. Games that require
        a network connection or remote authentication will fail to run from another account
        despite being installed. On the other hand, games that do not require network,
        can be executed, facilitating a rudimentary game sharing option on the same computer.

        @return bool If the game belongs to another count or not
        """
        ret = True
        try:
            if self.is_installed:
                _ = self.core.get_asset(self.game.app_name, platform=self.igame.platform).build_version
                ret = False
        except ValueError:
            logger.warning(f"Game {self.game.app_title} has no metadata. Set offline true")
        except AttributeError:
            ret = False
        return ret

    @property
    def needs_verification(self) -> bool:
        """!
        @brief Property to report if a games requires to be verified

        Simple wrapper around legendary's attribute with installation
        status check

        @return bool If the games needs to be verified
        """
        if self.igame is not None:
            return self.igame.needs_verification
        else:
            return False

    @needs_verification.setter
    def needs_verification(self, needs: bool) -> None:
        """!
        @brief Sets the verification status of a game.

        The operation here is reversed. since the property is
        named like this. After the verification, set this to 'False'
        to update the InstalledGame in the widget.

        @param needs If the game requires verification
        @return None
        """
        self.igame.needs_verification = needs
        self.store_igame()
        # FIXME: This might not be right to do for DLCs with actual data
        for dlc in self.owned_dlcs:
            if dlc.is_installed:
                dlc.needs_verification = needs
        if not needs:
            if not self.is_dlc:
                self.core.egstore_delete(self.igame)
            self.core.egstore_write(self.igame.app_name)

    @property
    def repair_file(self) -> str:
        return os.path.join(self.core.lgd.get_tmp_path(), f"{self.app_name}.repair")

    @property
    def needs_repair(self) -> bool:
        return os.path.exists(self.repair_file)

    @needs_repair.setter
    def needs_repair(self, needs: bool) -> None:
        if not needs and os.path.exists(self.repair_file):
            os.unlink(self.repair_file)

    @property
    def is_dlc(self) -> bool:
        """!
        @brief Property to report if Game is a dlc

        @return bool
        """
        return self.game.is_dlc

    @property
    def is_unreal(self) -> bool:
        """!
        @brief Property to report if a Game is an Unreal Engine bundle

        @return bool
        """
        return False if self.is_non_asset else self.game.asset_infos["Windows"].namespace == "ue"

    @property
    def is_non_asset(self) -> bool:
        """!
        @brief Property to report if a Game doesn't have assets

        Typically, games have assets, however some games that require
        other launchers do not have them. Rare treats these games as installed
        offering to execute their launcher.

        @return bool If the game doesn't have assets
        """

        # Asset infos are usually None, but there was a bug, that it was an empty GameAsset class
        return not self.game.asset_infos or not next(iter(self.game.asset_infos.values())).app_name

    @property
    def is_ubisoft(self) -> bool:
        return (
            self.game.metadata.get("customAttributes", {}).get("partnerLinkType", {}).get("value") == "ubisoft"
        )

    @property
    def folder_name(self) -> str:
        return (
            folder_name
            if (folder_name := self.game.metadata.get("customAttributes", {}).get("FolderName", {}).get("value"))
            else self.app_title
        )

    @property
    def sdl_name(self) -> Optional[str]:
        return get_sdl_appname(self.app_name)

    @property
    def save_path(self) -> Optional[str]:
        return super(RareGame, self).save_path

    @save_path.setter
    def save_path(self, path: str) -> None:
        if self.igame and (self.game.supports_cloud_saves or self.game.supports_mac_cloud_saves):
            self.igame.save_path = path
            self.store_igame()
            self.signals.widget.update.emit()

    def steam_grade(self) -> str:
        if platform.system() == "Windows" or self.is_unreal:
            return "na"
        elapsed_time = abs(datetime.utcnow() - self.metadata.steam_date)
        if self.metadata.steam_grade is not None and elapsed_time.days < 3:
            return self.metadata.steam_grade
        worker = QRunnable.create(
            lambda: self.set_steam_grade(get_rating(self.core, self.app_name))
        )
        QThreadPool.globalInstance().start(worker)
        return "pending"

    def set_steam_grade(self, grade: str) -> None:
        self.metadata.steam_grade = grade
        self.metadata.steam_date = datetime.utcnow()
        self.__save_metadata()
        self.signals.widget.update.emit()

    def grant_date(self, force=False) -> datetime:
        if self.metadata.grant_date is None or force:
            logger.debug("Grant date for %s not found in metadata, resolving", self.app_name)
            entitlements = self.core.lgd.entitlements
            matching = filter(lambda ent: ent["namespace"] == self.game.namespace, entitlements)
            entitlement = next(matching, None)
            grant_date = datetime.fromisoformat(
                entitlement["grantDate"].replace("Z", "+00:00")
            ) if entitlement else None
            self.metadata.grant_date = grant_date
            self.__save_metadata()
        return self.metadata.grant_date

    def set_origin_attributes(self, path: str, size: int = 0) -> None:
        self.__origin_install_path = path
        self.__origin_install_size = size
        if self.install_path and self.install_size:
            self.signals.game.installed.emit(self.app_name)
        else:
            self.signals.game.uninstalled.emit(self.app_name)
        self.set_pixmap()

    @property
    def can_launch(self) -> bool:
        if self.is_idle and self.is_origin:
            return True
        if self.is_installed:
            if (not self.is_idle) or self.needs_verification:
                return False
            if self.is_foreign and not self.can_run_offline:
                return False
            return True
        return False

    def set_pixmap(self):
        self.pixmap = self.image_manager.get_pixmap(self.app_name, self.is_installed)
        if not self.pixmap.isNull():
            self.signals.widget.update.emit()

    def load_pixmap(self):
        """ Do not call this function, call set_pixmap instead. This is only used for startup image loading """
        if self.pixmap.isNull():
            self.image_manager.download_image(self.game, self.set_pixmap, 0, False)

    def refresh_pixmap(self):
        self.image_manager.download_image(self.game, self.set_pixmap, 0, True)

    def install(self) -> bool:
        if not self.is_idle:
            return False
        self.signals.game.install.emit(
            InstallOptionsModel(app_name=self.app_name)
        )
        return True

    def modify(self) -> bool:
        if not self.is_idle:
            return False
        self.signals.game.install.emit(
            InstallOptionsModel(
                app_name=self.app_name, reset_sdl=True
            )
        )
        return True

    def repair(self, repair_and_update) -> bool:
        if not self.is_idle:
            return False
        self.signals.game.install.emit(
            InstallOptionsModel(
                app_name=self.app_name, repair_mode=True, repair_and_update=repair_and_update, update=repair_and_update
            )
        )
        return True

    def uninstall(self) -> bool:
        if not self.is_idle:
            return False
        self.signals.game.uninstall.emit(
            UninstallOptionsModel(app_name=self.app_name)
        )
        return True

    def launch(
        self,
        offline: bool = False,
        skip_update_check: bool = False,
        wine_bin: Optional[str] = None,
        wine_pfx: Optional[str] = None,
    ) -> bool:
        if not self.can_launch:
            return False

        cmd_line = get_rare_executable()
        executable, args = cmd_line[0], cmd_line[1:]
        args.extend(["launch", self.app_name])
        if offline:
            args.append("--offline")
        if skip_update_check:
            args.append("--skip-update-check")
        if wine_bin:
            args.extend(["--wine-bin", wine_bin])
        if wine_pfx:
            args.extend(["--wine-prefix", wine_pfx])

        QProcess.startDetached(executable, args)
        logger.info(f"Start new Process: ({executable} {' '.join(args)})")
        self.game_process.connect_to_server(on_startup=False)
        return True


class RareEosOverlay(RareGameBase):
    def __init__(self, legendary_core: LegendaryCore, game: Game):
        super(RareEosOverlay, self).__init__(legendary_core, game)
        self.igame: Optional[InstalledGame] = self.core.lgd.get_overlay_install_info()

    @property
    def is_installed(self) -> bool:
        return self.igame is not None

    def set_installed(self, installed: bool) -> None:
        if installed:
            self.igame = self.core.lgd.get_overlay_install_info()
            self.signals.game.installed.emit(self.app_name)
        else:
            self.igame = None
            self.signals.game.uninstalled.emit(self.app_name)



import os.path
import platform
from logging import getLogger
from typing import Tuple

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel, QFileDialog
from legendary.models.game import Game, InstalledGame

from rare.components.tabs.settings import DefaultGameSettings
from rare.components.tabs.settings.widgets.pre_launch import PreLaunchSettings
from rare.models.game import RareGame
from rare.utils import config_helper
from rare.widgets.side_tab import SideTabContents
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon

logger = getLogger("GameSettings")


class GameSettings(DefaultGameSettings, SideTabContents):
    def __init__(self, parent=None):
        super(GameSettings, self).__init__(False, parent=parent)
        self.pre_launch_settings = PreLaunchSettings()
        self.ui.launch_settings_group.layout().addRow(
            QLabel(self.tr("Pre-launch command")), self.pre_launch_settings
        )

        self.ui.skip_update.currentIndexChanged.connect(
            lambda x: self.update_combobox("skip_update_check", x)
        )
        self.ui.offline.currentIndexChanged.connect(
            lambda x: self.update_combobox("offline", x)
        )
        self.ui.launch_params.textChanged.connect(
            lambda x: self.line_edit_save_callback("start_params", x)
        )

        self.override_exe_edit = PathEdit(
            file_mode=QFileDialog.ExistingFile,
            name_filters=["*.exe", "*.app"],
            placeholder=self.tr("Relative path to launch executable"),
            edit_func=self.override_exe_edit_callback,
            save_func=self.override_exe_save_callback,
            parent=self
        )
        self.ui.launch_settings_layout.insertRow(
            self.ui.launch_settings_layout.getWidgetPosition(self.ui.launch_params)[0] + 1,
            QLabel(self.tr("Override executable"), self), self.override_exe_edit
        )

        self.ui.game_settings_layout.setAlignment(Qt.AlignTop)

        self.game: Game = None
        self.igame: InstalledGame = None

    def override_exe_edit_callback(self, path: str) -> Tuple[bool, str, int]:
        if not path or self.igame is None:
            return True, path, IndicatorReasonsCommon.VALID
        if not os.path.isabs(path):
            path = os.path.join(self.igame.install_path, path)
        if self.igame.install_path not in path:
            return False, self.igame.install_path, IndicatorReasonsCommon.WRONG_PATH
        if not os.path.exists(path):
            return False, path, IndicatorReasonsCommon.WRONG_PATH

        if not path.endswith(".exe") and not path.endswith(".app"):
            return False, path, IndicatorReasonsCommon.WRONG_PATH
        path = os.path.relpath(path, self.igame.install_path)
        return True, path, IndicatorReasonsCommon.VALID

    def override_exe_save_callback(self, path: str):
        self.line_edit_save_callback("override_exe", path)

    def line_edit_save_callback(self, option, value) -> None:
        if value:
            config_helper.add_option(self.game.app_name, option, value)
        else:
            config_helper.remove_option(self.game.app_name, option)
        config_helper.save_config()

    def update_combobox(self, option, index):
        if self.change:
            # remove section
            if index:
                if index == 1:
                    config_helper.add_option(self.game.app_name, option, "true")
                if index == 2:
                    config_helper.add_option(self.game.app_name, option, "false")
            else:
                config_helper.remove_option(self.game.app_name, option)
            config_helper.save_config()

    def load_settings(self, rgame: RareGame):
        self.change = False
        # FIXME: Use RareGame for the rest of the code
        app_name = rgame.app_name
        super(GameSettings, self).load_settings(app_name)
        self.game = rgame.game
        self.igame = rgame.igame
        if self.igame:
            if self.igame.can_run_offline:
                offline = self.core.lgd.config.get(self.game.app_name, "offline", fallback="unset")
                if offline == "true":
                    self.ui.offline.setCurrentIndex(1)
                elif offline == "false":
                    self.ui.offline.setCurrentIndex(2)
                else:
                    self.ui.offline.setCurrentIndex(0)

                self.ui.offline.setEnabled(True)
            else:
                self.ui.offline.setEnabled(False)
            self.override_exe_edit.set_root(self.igame.install_path)
        else:
            self.ui.offline.setEnabled(False)
            self.override_exe_edit.set_root("")

        skip_update = self.core.lgd.config.get(self.game.app_name, "skip_update_check", fallback="unset")
        if skip_update == "true":
            self.ui.skip_update.setCurrentIndex(1)
        elif skip_update == "false":
            self.ui.skip_update.setCurrentIndex(2)
        else:
            self.ui.skip_update.setCurrentIndex(0)

        self.set_title.emit(self.game.app_title)
        if platform.system() != "Windows":
            if self.igame and self.igame.platform == "Mac":
                self.ui.linux_settings_widget.setVisible(False)
            else:
                self.ui.linux_settings_widget.setVisible(True)

        self.ui.launch_params.setText(self.core.lgd.config.get(self.game.app_name, "start_params", fallback=""))
        self.override_exe_edit.setText(
            self.core.lgd.config.get(self.game.app_name, "override_exe", fallback="")
        )
        self.pre_launch_settings.load_settings(app_name)

        self.change = True



import multiprocessing
import os
import pathlib
import sys
from argparse import ArgumentParser


def main() -> int:
    # If we are on Windows, and we are in a "compiled" GUI application form
    # stdout (and stderr?) will be None. So to avoid `'NoneType' object has no attribute 'write'`
    # errors, redirect both of them to devnull
    if os.name == "nt" and (getattr(sys, "frozen", False) or ("__compiled__" in globals())):
        # Check if stdout and stderr are None before redirecting
        # This is useful in the case of test builds that enable console
        if sys.stdout is None:
            sys.stdout = open(os.devnull, 'w')
        if sys.stderr is None:
            sys.stderr = open(os.devnull, 'w')

    os.environ["QT_QPA_PLATFORMTHEME"] = ""
    if "LEGENDARY_CONFIG_PATH" in os.environ:
        os.environ["LEGENDARY_CONFIG_PATH"] = os.path.expanduser(os.environ["LEGENDARY_CONFIG_PATH"])

    # fix cx_freeze
    multiprocessing.freeze_support()

    # insert legendary for installed via pip/setup.py submodule to path
    # if not __name__ == "__main__":
    #     sys.path.insert(0, os.path.join(os.path.dirname(__file__), "legendary"))

    # CLI Options
    parser = ArgumentParser()
    parser.add_argument(
        "-V", "--version", action="store_true", help="Shows version and exits"
    )
    parser.add_argument(
        "-S",
        "--silent",
        action="store_true",
        help="Launch Rare in background. Open it from System Tray Icon",
    )
    parser.add_argument("--debug", action="store_true", help="Launch in debug mode")
    parser.add_argument(
        "--offline", action="store_true", help="Launch Rare in offline mode"
    )
    parser.add_argument(
        "--test-start", action="store_true", help="Quit immediately after launch"
    )

    parser.add_argument(
        "--desktop-shortcut",
        action="store_true",
        dest="desktop_shortcut",
        help="Use this, if there is no link on desktop to start Rare",
    )
    parser.add_argument(
        "--startmenu-shortcut",
        action="store_true",
        dest="startmenu_shortcut",
        help="Use this, if there is no link in start menu to launch Rare",
    )
    subparsers = parser.add_subparsers(title="Commands", dest="subparser")

    launch_minimal_parser = subparsers.add_parser("start", aliases=["launch"])
    launch_minimal_parser.add_argument("app_name", help="AppName of the game to launch",
                                       metavar="<App Name>", action="store")
    launch_minimal_parser.add_argument("--dry-run", help="Print arguments and exit", action="store_true")
    launch_minimal_parser.add_argument("--offline", help="Launch game offline",
                                       action="store_true")
    launch_minimal_parser.add_argument('--wine-bin', dest='wine_bin', action='store', metavar='<wine binary>',
                                       default=os.environ.get('LGDRY_WINE_BINARY', None),
                                       help='Set WINE binary to use to launch the app')
    launch_minimal_parser.add_argument('--wine-prefix', dest='wine_pfx', action='store', metavar='<wine pfx path>',
                                       default=os.environ.get('LGDRY_WINE_PREFIX', None),
                                       help='Set WINE prefix to use')
    launch_minimal_parser.add_argument("--ask-sync-saves", help="Ask to sync cloud saves",
                                       action="store_true")
    launch_minimal_parser.add_argument("--skip-update-check", help="Do not check for updates",
                                       action="store_true")

    args = parser.parse_args()

    if args.desktop_shortcut or args.startmenu_shortcut:
        from rare.utils.paths import create_desktop_link

        if args.desktop_shortcut:
            create_desktop_link(app_name="rare_shortcut", link_type="desktop")

        if args.startmenu_shortcut:
            create_desktop_link(app_name="rare_shortcut", link_type="start_menu")

        print("Link created")
        return 0

    if args.version:
        from rare import __version__, __codename__
        print(f"Rare {__version__} Codename: {__codename__}")
        return 0

    if args.subparser == "start" or args.subparser == "launch":
        from rare.launcher import launch
        return launch(args)

    from rare.utils import singleton

    try:
        # this object only allows one instance per machine

        me = singleton.SingleInstance()
    except singleton.SingleInstanceException:
        print("Rare is already running")
        from rare.utils.paths import lock_file

        with open(lock_file(), "w") as file:
            file.write("show")
            file.close()
        return -1

    from rare.components import start
    return start(args)


if __name__ == "__main__":
    # run from source
    # insert raw legendary submodule
    # sys.path.insert(
    #     0, os.path.join(pathlib.Path(__file__).parent.absolute(), "legendary")
    # )

    # insert source directory if running `main.py` as python script
    # Required by AppImage
    if "__compiled__" not in globals():
        sys.path.insert(0, str(pathlib.Path(__file__).parents[1].absolute()))

    sys.exit(main())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/proton.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ProtonSettings(object):
    def setupUi(self, ProtonSettings):
        ProtonSettings.setObjectName("ProtonSettings")
        ProtonSettings.resize(190, 86)
        ProtonSettings.setWindowTitle("ProtonSettings")
        self.proton_settings_layout = QtWidgets.QFormLayout(ProtonSettings)
        self.proton_settings_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.proton_settings_layout.setObjectName("proton_settings_layout")
        self.proton_wrapper_label = QtWidgets.QLabel(ProtonSettings)
        self.proton_wrapper_label.setObjectName("proton_wrapper_label")
        self.proton_settings_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.proton_wrapper_label)
        self.proton_combo = QtWidgets.QComboBox(ProtonSettings)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.proton_combo.sizePolicy().hasHeightForWidth())
        self.proton_combo.setSizePolicy(sizePolicy)
        self.proton_combo.setObjectName("proton_combo")
        self.proton_combo.addItem("")
        self.proton_settings_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.proton_combo)
        self.proton_prefix_label = QtWidgets.QLabel(ProtonSettings)
        self.proton_prefix_label.setObjectName("proton_prefix_label")
        self.proton_settings_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.proton_prefix_label)
        self.prefix_layout = QtWidgets.QHBoxLayout()
        self.prefix_layout.setObjectName("prefix_layout")
        self.proton_settings_layout.setLayout(1, QtWidgets.QFormLayout.FieldRole, self.prefix_layout)

        self.retranslateUi(ProtonSettings)

    def retranslateUi(self, ProtonSettings):
        _translate = QtCore.QCoreApplication.translate
        ProtonSettings.setTitle(_translate("ProtonSettings", "Proton Settings"))
        self.proton_wrapper_label.setText(_translate("ProtonSettings", "Proton"))
        self.proton_combo.setItemText(0, _translate("ProtonSettings", "Don\'t use Proton"))
        self.proton_prefix_label.setText(_translate("ProtonSettings", "Prefix"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ProtonSettings = QtWidgets.QGroupBox()
    ui = Ui_ProtonSettings()
    ui.setupUi(ProtonSettings)
    ProtonSettings.show()
    sys.exit(app.exec_())



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QHBoxLayout,
    QSpacerItem,
    QWidget,
)

from rare.utils.misc import icon
from rare.widgets.elide_label import ElideLabel


class ListWidget(object):
    def __init__(self):
        self.title_label = None
        self.status_label = None
        self.tooltip_label = None
        self.install_btn = None
        self.launch_btn = None

        self.developer_label = None
        self.version_label = None
        self.size_label = None

    def setupUi(self, widget: QWidget):
        self.title_label = QLabel(parent=widget)
        self.title_label.setObjectName(f"{type(self).__name__}TitleLabel")
        self.title_label.setWordWrap(False)

        self.status_label = QLabel(parent=widget)
        self.status_label.setObjectName(f"{type(self).__name__}StatusLabel")
        self.status_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        self.tooltip_label = QLabel(parent=widget)
        self.tooltip_label.setObjectName(f"{type(self).__name__}TooltipLabel")
        self.tooltip_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        self.install_btn = QPushButton(parent=widget)
        self.install_btn.setObjectName(f"{type(self).__name__}Button")
        self.install_btn.setIcon(icon("ri.install-fill"))
        self.install_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.install_btn.setFixedWidth(120)

        self.launch_btn = QPushButton(parent=widget)
        self.launch_btn.setObjectName(f"{type(self).__name__}Button")
        self.launch_btn.setIcon(icon("ei.play-alt"))
        self.launch_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.launch_btn.setFixedWidth(120)

        # lk: do not focus on button
        # When the button gets clicked on, it receives keyboard focus. Disabling the button
        # afterwards leads to `focusNextChild` getting called. This makes the scrollarea
        # trying to ensure that `nextChild` is visible, essentially scrolling to a random widget
        self.launch_btn.setFocusPolicy(Qt.NoFocus)
        self.install_btn.setFocusPolicy(Qt.NoFocus)

        self.developer_label = ElideLabel(parent=widget)
        self.developer_label.setObjectName(f"{type(self).__name__}InfoLabel")
        self.developer_label.setFixedWidth(120)

        self.version_label = ElideLabel(parent=widget)
        self.version_label.setObjectName(f"{type(self).__name__}InfoLabel")
        self.version_label.setFixedWidth(120)

        self.size_label = ElideLabel(parent=widget)
        self.size_label.setObjectName(f"{type(self).__name__}InfoLabel")
        self.size_label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.size_label.setFixedWidth(60)

        # Create layouts
        top_layout = QHBoxLayout()
        top_layout.setAlignment(Qt.AlignLeft)

        bottom_layout = QHBoxLayout()
        bottom_layout.setAlignment(Qt.AlignRight)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(3, 3, 3, 3)

        # Layout the widgets
        # (from inner to outer)
        top_layout.addWidget(self.title_label, stretch=1)

        bottom_layout.addWidget(self.developer_label, stretch=0, alignment=Qt.AlignLeft)
        bottom_layout.addItem(QSpacerItem(20, 0, QSizePolicy.Fixed, QSizePolicy.Minimum))
        bottom_layout.addWidget(self.version_label, stretch=0, alignment=Qt.AlignLeft)
        bottom_layout.addItem(QSpacerItem(20, 0, QSizePolicy.Fixed, QSizePolicy.Minimum))
        bottom_layout.addWidget(self.size_label, stretch=0, alignment=Qt.AlignLeft)
        bottom_layout.addItem(QSpacerItem(20, 0, QSizePolicy.Fixed, QSizePolicy.Minimum))
        bottom_layout.addWidget(self.status_label, stretch=0, alignment=Qt.AlignLeft)
        bottom_layout.addItem(QSpacerItem(20, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        bottom_layout.addWidget(self.tooltip_label, stretch=0, alignment=Qt.AlignRight)
        bottom_layout.addWidget(self.install_btn, stretch=0, alignment=Qt.AlignRight)
        bottom_layout.addWidget(self.launch_btn, stretch=0, alignment=Qt.AlignRight)

        layout.addLayout(top_layout)
        layout.addLayout(bottom_layout)

        widget.setLayout(layout)

        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        widget.leaveEvent(None)

        self.translateUi(widget)

    def translateUi(self, widget: QWidget):
        self.install_btn.setText(widget.tr("Install"))



import os
from logging import getLogger
from pathlib import Path
from typing import Tuple

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QGroupBox, QFileDialog

from rare.components.tabs.settings import LinuxSettings
from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.settings.proton import Ui_ProtonSettings
from rare.utils import config_helper
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon
from .wrapper import WrapperSettings

logger = getLogger("Proton")


def find_proton_combos():
    possible_proton_combos = []
    compatibilitytools_dirs = [
        os.path.expanduser("~/.steam/steam/steamapps/common"),
        "/usr/share/steam/compatibilitytools.d",
        os.path.expanduser("~/.steam/compatibilitytools.d"),
        os.path.expanduser("~/.steam/root/compatibilitytools.d"),
    ]
    for c in compatibilitytools_dirs:
        if os.path.exists(c):
            for i in os.listdir(c):
                proton = os.path.join(c, i, "proton")
                compatibilitytool = os.path.join(c, i, "compatibilitytool.vdf")
                toolmanifest = os.path.join(c, i, "toolmanifest.vdf")
                if os.path.exists(proton) and (
                        os.path.exists(compatibilitytool) or os.path.exists(toolmanifest)
                ):
                    wrapper = f'"{proton}" run'
                    possible_proton_combos.append(wrapper)
    if not possible_proton_combos:
        logger.warning("Unable to find any Proton version")
    return possible_proton_combos


class ProtonSettings(QGroupBox):
    # str: option key
    environ_changed = pyqtSignal(str)
    app_name: str
    changeable = True

    def __init__(self, linux_settings: LinuxSettings, wrapper_settings: WrapperSettings):
        super(ProtonSettings, self).__init__()
        self.ui = Ui_ProtonSettings()
        self.ui.setupUi(self)
        self._linux_settings = linux_settings
        self._wrapper_settings = wrapper_settings
        self.core = LegendaryCoreSingleton()
        self.possible_proton_combos = find_proton_combos()

        self.ui.proton_combo.addItems(self.possible_proton_combos)
        self.ui.proton_combo.currentIndexChanged.connect(self.change_proton)

        self.proton_prefix = PathEdit(
            file_mode=QFileDialog.DirectoryOnly,
            edit_func=self.proton_prefix_edit,
            save_func=self.proton_prefix_save,
            placeholder=self.tr("Please select path for proton prefix")
        )
        self.ui.prefix_layout.addWidget(self.proton_prefix)

    def change_proton(self, i):
        if not self.changeable:
            return
        # First combo box entry: Don't use Proton
        if i == 0:
            self._wrapper_settings.delete_wrapper("proton")
            config_helper.remove_option(self.app_name, "no_wine")
            config_helper.remove_option(f"{self.app_name}.env", "STEAM_COMPAT_DATA_PATH")
            self.environ_changed.emit("STEAM_COMPAT_DATA_PATH")
            config_helper.remove_option(f"{self.app_name}.env", "STEAM_COMPAT_CLIENT_INSTALL_PATH")
            self.environ_changed.emit("STEAM_COMPAT_CLIENT_INSTALL_PATH")

            self.proton_prefix.setEnabled(False)
            self.proton_prefix.setText("")

            self._linux_settings.ui.wine_groupbox.setEnabled(True)
        else:
            self.proton_prefix.setEnabled(True)
            self._linux_settings.ui.wine_groupbox.setEnabled(False)
            wrapper = self.possible_proton_combos[i - 1]
            self._wrapper_settings.add_wrapper(wrapper)
            config_helper.add_option(self.app_name, "no_wine", "true")
            config_helper.add_option(
                f"{self.app_name}.env",
                "STEAM_COMPAT_CLIENT_INSTALL_PATH",
                str(Path.home().joinpath(".steam", "steam"))
            )
            self.environ_changed.emit("STEAM_COMPAT_CLIENT_INSTALL_PATH")

            self.proton_prefix.setText(os.path.expanduser("~/.proton"))

            # Don't use Wine
            self._linux_settings.wine_exec.setText("")
            self._linux_settings.wine_prefix.setText("")

        config_helper.save_config()

    def proton_prefix_edit(self, text: str) -> Tuple[bool, str, int]:
        if not text:
            return False, text, IndicatorReasonsCommon.EMPTY
        parent_dir = os.path.dirname(text)
        return os.path.exists(parent_dir), text, IndicatorReasonsCommon.DIR_NOT_EXISTS

    def proton_prefix_save(self, text: str):
        if not text:
            return
        config_helper.add_option(
            f"{self.app_name}.env", "STEAM_COMPAT_DATA_PATH", text
        )
        self.environ_changed.emit("STEAM_COMPAT_DATA_PATH")
        config_helper.save_config()

    def load_settings(self, app_name: str, proton: str):
        self.changeable = False
        self.app_name = app_name
        proton = proton.replace('"', "")
        self.proton_prefix.setEnabled(bool(proton))
        if proton:
            self.ui.proton_combo.setCurrentText(
                f'"{proton.replace(" run", "")}" run'
            )
        else:
            self.ui.proton_combo.setCurrentIndex(0)

        proton_prefix = self.core.lgd.config.get(
            f"{app_name}.env",
            "STEAM_COMPAT_DATA_PATH",
            fallback="",
        )
        self.proton_prefix.setText(proton_prefix)
        self.changeable = True



import os
import platform
import queue
import time
from dataclasses import dataclass
from enum import IntEnum
from logging import getLogger
from typing import List, Optional, Dict

from PyQt5.QtCore import QThread, pyqtSignal, QProcess

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.core import LegendaryCore
from rare.lgndr.glue.monkeys import DLManagerSignals
from rare.lgndr.models.downloading import UIUpdate
from rare.models.game import RareGame
from rare.models.install import InstallQueueItemModel, InstallOptionsModel

logger = getLogger("DownloadThread")


class DlResultCode(IntEnum):
    ERROR = 1
    STOPPED = 2
    FINISHED = 3


@dataclass
class DlResultModel:
    options: InstallOptionsModel
    code: DlResultCode = DlResultCode.ERROR
    message: str = ""
    dlcs: Optional[List[Dict]] = None
    sync_saves: bool = False
    tip_url: str = ""
    shortcut: bool = False
    shortcut_name: str = ""
    shortcut_title: str = ""


class DlThread(QThread):
    result = pyqtSignal(DlResultModel)
    progress = pyqtSignal(UIUpdate, object)

    def __init__(self, item: InstallQueueItemModel, rgame: RareGame, core: LegendaryCore, debug: bool = False):
        super(DlThread, self).__init__()
        self.dlm_signals: DLManagerSignals = DLManagerSignals()
        self.core: LegendaryCore = core
        self.item: InstallQueueItemModel = item
        self.dl_size = item.download.analysis.dl_size
        self.rgame = rgame
        self.debug = debug

    def __finish(self, result):
        if result.code == DlResultCode.FINISHED:
            self.rgame.set_installed(True)
        self.rgame.state = RareGame.State.IDLE
        self.rgame.signals.progress.finish.emit(not result.code == DlResultCode.FINISHED)
        self.result.emit(result)

    def __status_callback(self, status: UIUpdate):
        self.progress.emit(status, self.dl_size)
        self.rgame.signals.progress.update.emit(int(status.progress))

    def run(self):
        cli = LegendaryCLI(self.core)
        self.item.download.dlm.logging_queue = cli.logging_queue
        self.item.download.dlm.proc_debug = self.debug
        result = DlResultModel(self.item.options)
        start_t = time.time()
        try:
            self.item.download.dlm.start()
            self.rgame.state = RareGame.State.DOWNLOADING
            self.rgame.signals.progress.start.emit()
            time.sleep(1)
            while self.item.download.dlm.is_alive():
                try:
                    self.__status_callback(self.item.download.dlm.status_queue.get(timeout=1.0))
                except queue.Empty:
                    pass
                if self.dlm_signals.update:
                    try:
                        self.item.download.dlm.signals_queue.put(self.dlm_signals, block=False, timeout=1.0)
                    except queue.Full:
                        pass
                time.sleep(self.item.download.dlm.update_interval / 10)
            self.item.download.dlm.join()
        except Exception as e:
            self.kill()
            self.item.download.dlm.join()
            end_t = time.time()
            logger.error(f"Installation failed after {end_t - start_t:.02f} seconds.")
            logger.warning(f"The following exception occurred while waiting for the downloader to finish: {e!r}.")
            result.code = DlResultCode.ERROR
            result.message = f"{e!r}"
            self.__finish(result)
            return
        else:
            end_t = time.time()
            if self.dlm_signals.kill is True:
                logger.info(f"Download stopped after {end_t - start_t:.02f} seconds.")
                result.code = DlResultCode.STOPPED
                self.__finish(result)
                return
            logger.info(f"Download finished in {end_t - start_t:.02f} seconds.")

            result.code = DlResultCode.FINISHED

            if self.item.options.overlay:
                self.core.finish_overlay_install(self.item.download.igame)
                self.__finish(result)
                return

            if not self.item.options.no_install:
                postinstall = self.core.install_game(self.item.download.igame)
                if postinstall:
                    # LegendaryCLI(self.core)._handle_postinstall(
                    #     postinstall,
                    #     self.item.download.igame,
                    #     False,
                    #     self.item.options.install_prereqs,
                    # )
                    self._handle_postinstall(postinstall, self.item.download.igame)

                dlcs = self.core.get_dlc_for_game(self.item.download.igame.app_name)
                if dlcs and not self.item.options.skip_dlcs:
                    result.dlcs = []
                    for dlc in dlcs:
                        result.dlcs.append(
                            {
                                "app_name": dlc.app_name,
                                "app_title": dlc.app_title,
                                "app_version": dlc.app_version(self.item.options.platform),
                            }
                        )

                if (
                    self.item.download.game.supports_cloud_saves
                    or self.item.download.game.supports_mac_cloud_saves
                ) and not self.item.download.game.is_dlc:
                    result.sync_saves = True

                # show tip again after installation finishes so users hopefully actually see it
                if tip_url := self.core.get_game_tip(self.item.download.igame.app_name):
                    result.tip_url = tip_url

            LegendaryCLI(self.core).install_game_cleanup(
                self.item.download.game,
                self.item.download.igame,
                self.item.download.repair,
                self.item.download.repair_file,
            )

            if not self.item.options.update and self.item.options.create_shortcut:
                result.shortcut = True
                result.shortcut_name = self.rgame.folder_name
                result.shortcut_title = self.rgame.app_title

        self.__finish(result)

    def _handle_postinstall(self, postinstall, igame):
        logger.info("This game lists the following prerequisites to be installed:")
        logger.info(f'- {postinstall["name"]}: {" ".join((postinstall["path"], postinstall["args"]))}')
        if platform.system() == "Windows":
            if self.item.options.install_prereqs:
                logger.info("Launching prerequisite executable..")
                self.core.prereq_installed(igame.app_name)
                req_path, req_exec = os.path.split(postinstall["path"])
                work_dir = os.path.join(igame.install_path, req_path)
                fullpath = os.path.join(work_dir, req_exec)
                proc = QProcess()
                proc.setProcessChannelMode(QProcess.MergedChannels)
                proc.readyReadStandardOutput.connect(
                    lambda: logger.debug(str(proc.readAllStandardOutput().data(), "utf-8", "ignore"))
                )
                proc.setProgram(fullpath)
                proc.setArguments(postinstall.get("args", "").split(" "))
                proc.setWorkingDirectory(work_dir)
                proc.start()
                proc.waitForFinished()  # wait, because it is inside the thread
        else:
            logger.info("Automatic installation not available on Linux.")

    def kill(self):
        self.dlm_signals.kill = True



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/login/import_login.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ImportLogin(object):
    def setupUi(self, ImportLogin):
        ImportLogin.setObjectName("ImportLogin")
        ImportLogin.resize(233, 156)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ImportLogin.sizePolicy().hasHeightForWidth())
        ImportLogin.setSizePolicy(sizePolicy)
        ImportLogin.setWindowTitle("ImportLogin")
        self.main_layout = QtWidgets.QVBoxLayout(ImportLogin)
        self.main_layout.setObjectName("main_layout")
        self.title_label = QtWidgets.QLabel(ImportLogin)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.title_label.sizePolicy().hasHeightForWidth())
        self.title_label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.title_label.setFont(font)
        self.title_label.setObjectName("title_label")
        self.main_layout.addWidget(self.title_label)
        self.form_layout = QtWidgets.QFormLayout()
        self.form_layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.form_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.form_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.form_layout.setFormAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.form_layout.setObjectName("form_layout")
        self.prefix_layout = QtWidgets.QHBoxLayout()
        self.prefix_layout.setObjectName("prefix_layout")
        self.prefix_combo = QtWidgets.QComboBox(ImportLogin)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.prefix_combo.sizePolicy().hasHeightForWidth())
        self.prefix_combo.setSizePolicy(sizePolicy)
        self.prefix_combo.setEditable(True)
        self.prefix_combo.setObjectName("prefix_combo")
        self.prefix_layout.addWidget(self.prefix_combo)
        self.prefix_tool = QtWidgets.QToolButton(ImportLogin)
        self.prefix_tool.setObjectName("prefix_tool")
        self.prefix_layout.addWidget(self.prefix_tool)
        self.prefix_layout.setStretch(0, 1)
        self.form_layout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.prefix_layout)
        self.prefix_label = QtWidgets.QLabel(ImportLogin)
        self.prefix_label.setObjectName("prefix_label")
        self.form_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.prefix_label)
        self.status_label = QtWidgets.QLabel(ImportLogin)
        font = QtGui.QFont()
        font.setItalic(True)
        self.status_label.setFont(font)
        self.status_label.setText("")
        self.status_label.setObjectName("status_label")
        self.form_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.status_label)
        self.main_layout.addLayout(self.form_layout)
        spacerItem = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.main_layout.addItem(spacerItem)
        self.info_label = QtWidgets.QLabel(ImportLogin)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.info_label.sizePolicy().hasHeightForWidth())
        self.info_label.setSizePolicy(sizePolicy)
        self.info_label.setWordWrap(True)
        self.info_label.setObjectName("info_label")
        self.main_layout.addWidget(self.info_label)

        self.retranslateUi(ImportLogin)

    def retranslateUi(self, ImportLogin):
        _translate = QtCore.QCoreApplication.translate
        self.title_label.setText(_translate("ImportLogin", "Import existing session from EGL"))
        self.prefix_tool.setText(_translate("ImportLogin", "Browse"))
        self.prefix_label.setText(_translate("ImportLogin", "Select prefix"))
        self.info_label.setText(_translate("ImportLogin", "<i>Please select the Wine prefix where Epic Games Launcher is installed. You will get logged out from EGL in the process.</i>"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ImportLogin = QtWidgets.QWidget()
    ui = Ui_ImportLogin()
    ui.setupUi(ImportLogin)
    ImportLogin.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/import_sync/egl_sync_list_group.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EGLSyncListGroup(object):
    def setupUi(self, EGLSyncListGroup):
        EGLSyncListGroup.setObjectName("EGLSyncListGroup")
        EGLSyncListGroup.resize(461, 206)
        EGLSyncListGroup.setWindowTitle("EGLSyncListGroup")
        EGLSyncListGroup.setTitle("")
        self.egl_sync_list_layout = QtWidgets.QVBoxLayout(EGLSyncListGroup)
        self.egl_sync_list_layout.setObjectName("egl_sync_list_layout")
        self.label = QtWidgets.QLabel(EGLSyncListGroup)
        self.label.setText("")
        self.label.setObjectName("label")
        self.egl_sync_list_layout.addWidget(self.label)
        self.list = QtWidgets.QListWidget(EGLSyncListGroup)
        self.list.setAlternatingRowColors(True)
        self.list.setObjectName("list")
        self.egl_sync_list_layout.addWidget(self.list)
        self.buttons_widget = QtWidgets.QWidget(EGLSyncListGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttons_widget.sizePolicy().hasHeightForWidth())
        self.buttons_widget.setSizePolicy(sizePolicy)
        self.buttons_widget.setObjectName("buttons_widget")
        self.buttons_layout = QtWidgets.QHBoxLayout(self.buttons_widget)
        self.buttons_layout.setContentsMargins(0, 0, 0, 0)
        self.buttons_layout.setObjectName("buttons_layout")
        self.select_all_button = QtWidgets.QPushButton(self.buttons_widget)
        self.select_all_button.setObjectName("select_all_button")
        self.buttons_layout.addWidget(self.select_all_button)
        self.select_none_button = QtWidgets.QPushButton(self.buttons_widget)
        self.select_none_button.setObjectName("select_none_button")
        self.buttons_layout.addWidget(self.select_none_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.buttons_layout.addItem(spacerItem)
        self.action_button = QtWidgets.QPushButton(self.buttons_widget)
        self.action_button.setText("Action")
        self.action_button.setObjectName("action_button")
        self.buttons_layout.addWidget(self.action_button)
        self.egl_sync_list_layout.addWidget(self.buttons_widget)

        self.retranslateUi(EGLSyncListGroup)

    def retranslateUi(self, EGLSyncListGroup):
        _translate = QtCore.QCoreApplication.translate
        self.list.setSortingEnabled(True)
        self.select_all_button.setText(_translate("EGLSyncListGroup", "Select all"))
        self.select_none_button.setText(_translate("EGLSyncListGroup", "Select none"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EGLSyncListGroup = QtWidgets.QGroupBox()
    ui = Ui_EGLSyncListGroup()
    ui.setupUi(EGLSyncListGroup)
    EGLSyncListGroup.show()
    sys.exit(app.exec_())



import datetime
import sys
from logging import getLogger

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QSizePolicy, QLayout, QApplication, QWidget
from legendary.core import LegendaryCore
from legendary.models.game import InstalledGame

from rare.ui.components.dialogs.sync_save_dialog import Ui_SyncSaveDialog
from rare.ui.components.tabs.games.game_info.sync_widget import Ui_SyncWidget
from rare.utils.misc import icon

logger = getLogger("Cloud Saves")


class CloudSaveDialog(QDialog, Ui_SyncSaveDialog):
    DOWNLOAD = 2
    UPLOAD = 1
    CANCEL = 0
    SKIP = 3

    def __init__(
        self,
        igame: InstalledGame,
        dt_local: datetime.datetime,
        dt_remote: datetime.datetime,
    ):
        super(CloudSaveDialog, self).__init__()
        self.setupUi(self)

        self.sync_widget = QWidget()
        self.sync_ui = Ui_SyncWidget()
        self.sync_ui.setupUi(self.sync_widget)

        self.sync_widget_layout.addWidget(self.sync_widget)

        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint)

        self.status = self.CANCEL

        self.title_label.setText(f"{self.title_label.text()} <b>{igame.title}</b>")

        newer = self.tr("Newer")
        if dt_remote and dt_local:
            self.sync_ui.age_label_local.setText(
                f"<b>{newer}</b>" if dt_remote < dt_local else " "
            )
            self.sync_ui.age_label_remote.setText(
                f"<b>{newer}</b>" if dt_remote > dt_local else " "
            )
        # Set status, if one of them is None
        elif dt_remote and not dt_local:
            self.status = self.DOWNLOAD
        elif not dt_remote and dt_local:
            self.status = self.UPLOAD
        else:
            self.status = self.SKIP

        self.sync_ui.date_info_local.setText(dt_local.strftime("%A, %d. %B %Y %X") if dt_local else "None")
        self.sync_ui.date_info_remote.setText(dt_remote.strftime("%A, %d. %B %Y %X") if dt_remote else "None")

        self.sync_ui.icon_local.setPixmap(icon("mdi.harddisk", "fa.desktop").pixmap(128, 128))
        self.sync_ui.icon_remote.setPixmap(icon("mdi.cloud-outline", "ei.cloud").pixmap(128, 128))

        self.sync_ui.upload_button.clicked.connect(lambda: self.btn_clicked(self.UPLOAD))
        self.sync_ui.download_button.clicked.connect(lambda: self.btn_clicked(self.DOWNLOAD))
        self.cancel_button.clicked.connect(self.close)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def get_action(self):
        if self.status == self.SKIP:
            return self.SKIP
        self.exec_()
        return self.status

    def btn_clicked(self, status):
        self.status = status
        self.close()


def test_dialog():
    app = QApplication(sys.argv)
    core = LegendaryCore()
    dlg = CloudSaveDialog(core.get_installed_list()[0], datetime.datetime.now(),
                          datetime.datetime.strptime("2021,1", "%Y,%M"))
    print(dlg.get_action())


if __name__ == '__main__':
    test_dialog()



import logging

from PyQt5 import QtGui
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QFont
from PyQt5.QtNetwork import QNetworkAccessManager
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QHBoxLayout

from rare.components.tabs.store.shop_models import ImageUrlModel
from rare.ui.components.tabs.store.wishlist_widget import Ui_WishlistWidget
from rare.utils.extra_widgets import ImageLabel
from rare.utils.misc import icon

logger = logging.getLogger("GameWidgets")


class GameWidget(QWidget):
    show_info = pyqtSignal(dict)

    def __init__(self, path, json_info=None, width=300):
        super(GameWidget, self).__init__()
        self.manager = QNetworkAccessManager()
        self.width = width
        self.path = path
        if json_info:
            self.init_ui(json_info)

    def init_ui(self, json_info):
        self.layout = QVBoxLayout()
        self.image = ImageLabel()
        self.layout.addWidget(self.image)
        mini_layout = QHBoxLayout()
        self.layout.addLayout(mini_layout)

        if not json_info:
            self.layout.addWidget(QLabel("An error occurred"))
            self.setLayout(self.layout)
            return

        self.title_label = QLabel(json_info.get("title"))
        self.title_label.setWordWrap(True)
        mini_layout.addWidget(self.title_label)
        mini_layout.addStretch(1)

        price = json_info["price"]["totalPrice"]["fmtPrice"]["originalPrice"]
        discount_price = json_info["price"]["totalPrice"]["fmtPrice"]["discountPrice"]
        price_label = QLabel(price)
        if price != discount_price:
            font = QFont()
            font.setStrikeOut(True)
            price_label.setFont(font)
            mini_layout.addWidget(
                QLabel(discount_price if discount_price != "0" else self.tr("Free"))
            )
            mini_layout.addWidget(price_label)
        else:
            if price == "0":
                price_label.setText(self.tr("Free"))
            mini_layout.addWidget(price_label)

        for c in r'<>?":|\/*':
            json_info["title"] = json_info["title"].replace(c, "")

        self.json_info = json_info
        self.slug = json_info["productSlug"]

        self.title = json_info["title"]
        for img in json_info["keyImages"]:
            if img["type"] in [
                "DieselStoreFrontWide",
                "OfferImageWide",
                "VaultClosed",
                "ProductLogo",
            ]:
                if img["type"] == "VaultClosed" and self.title != "Mystery Game":
                    continue
                self.image.update_image(
                    img["url"],
                    json_info["title"],
                    (self.width, int(self.width * 9 / 16)),
                )
                break
        else:
            logger.info(", ".join([img["type"] for img in json_info["keyImages"]]))

        self.setLayout(self.layout)

        self.setFixedSize(self.width + 10, self.width * 9 // 16 + 50)

    def mousePressEvent(self, a0: QtGui.QMouseEvent) -> None:
        self.show_info.emit(self.json_info)


class WishlistWidget(QWidget, Ui_WishlistWidget):
    open_game = pyqtSignal(dict)
    delete_from_wishlist = pyqtSignal(dict)

    def __init__(self, game: dict):
        super(WishlistWidget, self).__init__()
        self.setupUi(self)
        self.game = game
        self.title_label.setText(game.get("title"))
        for attr in game["customAttributes"]:
            if attr["key"] == "developerName":
                self.developer.setText(attr["value"])
                break
        else:
            self.developer.setText(game["seller"]["name"])
        original_price = game["price"]["totalPrice"]["fmtPrice"]["originalPrice"]
        discount_price = game["price"]["totalPrice"]["fmtPrice"]["discountPrice"]

        self.price.setText(original_price if original_price != "0" else self.tr("Free"))
        # if discount
        if original_price != discount_price:
            self.discount = True
            font = QFont()
            font.setStrikeOut(True)
            self.price.setFont(font)
            self.discount_price.setText(discount_price)
        else:
            self.discount = False
            self.discount_price.setVisible(False)

        self.image = ImageLabel()
        self.layout().insertWidget(0, self.image)
        image_model = ImageUrlModel.from_json(game["keyImages"])
        url = image_model.front_wide
        if not url:
            url = image_model.offer_image_wide
        self.image.update_image(url, game.get("title"), (240, 135))
        self.delete_button.setIcon(icon("mdi.delete", color="white"))
        self.delete_button.clicked.connect(
            lambda: self.delete_from_wishlist.emit(self.game)
        )

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        # left button
        if e.button() == 1:
            self.open_game.emit(self.game)
        # right
        elif e.button() == 2:
            pass  # self.showMenu(e)



from logging import getLogger
from typing import Union, Protocol

from PyQt5.QtCore import (
    Qt,
    QRect,
    QSize,
    QPoint,
    pyqtSignal,
)
from PyQt5.QtGui import QFontMetrics
from PyQt5.QtWidgets import (
    QStyle,
    QLabel,
    QWidget,
    QStyleOptionTab,
    QStylePainter,
    QTabBar,
    QTabWidget,
    QVBoxLayout,
    QScrollArea, QLayout, QSizePolicy,
)

from rare.utils.misc import icon as qta_icon

logger = getLogger("SideTab")


class SideTabBar(QTabBar):
    def __init__(self, padding: int = -1, parent=None):
        super(SideTabBar, self).__init__(parent=parent)
        self.setObjectName("SideTabBar")
        self.padding = padding
        self.fm = QFontMetrics(self.font())

    def tabSizeHint(self, index):
        width = QTabBar.tabSizeHint(self, index).height()
        if self.padding < 0:
            width += QTabBar.tabSizeHint(self, index).width()
        else:
            width += self.padding
        return QSize(width, self.fm.height() + 18)

    def paintEvent(self, event):
        painter = QStylePainter(self)
        opt = QStyleOptionTab()

        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(QStyle.CE_TabBarTabShape, opt)
            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = QRect(QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            c = self.tabRect(i).center()
            painter.translate(c)
            painter.rotate(90)
            painter.translate(-c)
            painter.drawControl(QStyle.CE_TabBarTabLabel, opt)
            painter.restore()


class SideTabContents(object):
    # str: title
    set_title = pyqtSignal(str)
    implements_scrollarea: bool = False


class SideTabContentsProtocol(Protocol):
    implements_scrollarea: bool

    def layout(self) -> QLayout:
        pass

    def set_title(self) -> pyqtSignal:
        pass

    def sizeHint(self) -> QSize:
        pass


class SideTabContainer(QWidget):
    def __init__(self, widget: Union[QWidget, SideTabContentsProtocol], title: str = "", parent: QWidget = None):
        super(SideTabContainer, self).__init__(parent=parent)
        self.title = QLabel(self)
        self.setTitle(title)

        if widget.layout():
            widget.layout().setAlignment(Qt.AlignTop)
            widget.layout().setContentsMargins(0, 0, 3, 0)
        if hasattr(widget, "set_title"):
            widget.set_title.connect(self.setTitle)

        layout = QVBoxLayout(self)
        layout.addWidget(self.title)

        if not hasattr(widget, "implements_scrollarea") or not widget.implements_scrollarea:
            self.scrollarea = QScrollArea(self)
            self.scrollarea.setSizeAdjustPolicy(QScrollArea.AdjustToContents)
            self.scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            self.scrollarea.setFrameStyle(QScrollArea.NoFrame)
            self.scrollarea.setMinimumWidth(
                widget.sizeHint().width() + self.scrollarea.verticalScrollBar().sizeHint().width()
            )
            self.scrollarea.setWidgetResizable(True)
            widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
            self.scrollarea.setWidget(widget)
            layout.addWidget(self.scrollarea)
        else:
            layout.addWidget(widget)

    def setTitle(self, text: str) -> None:
        self.title.setText(f"<h2>{text}</h2>")
        self.title.setVisible(bool(text))


class SideTabWidget(QTabWidget):
    back_clicked = pyqtSignal()

    def __init__(self, show_back: bool = False, padding: int = -1, parent=None):
        super(SideTabWidget, self).__init__(parent=parent)
        self.setTabBar(SideTabBar(padding=padding, parent=self))
        self.setDocumentMode(True)
        self.setTabPosition(QTabWidget.West)
        if show_back:
            super(SideTabWidget, self).addTab(
                QWidget(self), qta_icon("mdi.keyboard-backspace", "ei.backward"), self.tr("Back")
            )
            self.tabBarClicked.connect(self.back_func)

    def back_func(self, tab):
        # shortcut for tab == 0
        if not tab:
            self.back_clicked.emit()

    def addTab(self, widget: Union[QWidget, SideTabContentsProtocol], a1: str, title: str = "") -> int:
        container = SideTabContainer(widget, title, parent=self)
        return super(SideTabWidget, self).addTab(container, a1)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/login/landing_page.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LandingPage(object):
    def setupUi(self, LandingPage):
        LandingPage.setObjectName("LandingPage")
        LandingPage.resize(293, 78)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LandingPage.sizePolicy().hasHeightForWidth())
        LandingPage.setSizePolicy(sizePolicy)
        LandingPage.setWindowTitle("LandingPage")
        self.main_layout = QtWidgets.QFormLayout(LandingPage)
        self.main_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.main_layout.setLabelAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.main_layout.setFormAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.main_layout.setObjectName("main_layout")
        self.login_label = QtWidgets.QLabel(LandingPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.login_label.sizePolicy().hasHeightForWidth())
        self.login_label.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.login_label.setFont(font)
        self.login_label.setObjectName("login_label")
        self.main_layout.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.login_label)
        self.login_browser_radio = QtWidgets.QRadioButton(LandingPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.login_browser_radio.sizePolicy().hasHeightForWidth())
        self.login_browser_radio.setSizePolicy(sizePolicy)
        self.login_browser_radio.setObjectName("login_browser_radio")
        self.main_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.login_browser_radio)
        self.login_browser_label = QtWidgets.QLabel(LandingPage)
        font = QtGui.QFont()
        font.setItalic(True)
        self.login_browser_label.setFont(font)
        self.login_browser_label.setObjectName("login_browser_label")
        self.main_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.login_browser_label)
        self.login_import_radio = QtWidgets.QRadioButton(LandingPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.login_import_radio.sizePolicy().hasHeightForWidth())
        self.login_import_radio.setSizePolicy(sizePolicy)
        self.login_import_radio.setObjectName("login_import_radio")
        self.main_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.login_import_radio)
        self.login_import_label = QtWidgets.QLabel(LandingPage)
        font = QtGui.QFont()
        font.setItalic(True)
        self.login_import_label.setFont(font)
        self.login_import_label.setObjectName("login_import_label")
        self.main_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.login_import_label)

        self.retranslateUi(LandingPage)

    def retranslateUi(self, LandingPage):
        _translate = QtCore.QCoreApplication.translate
        self.login_label.setText(_translate("LandingPage", "Select login method"))
        self.login_browser_radio.setText(_translate("LandingPage", "Browser"))
        self.login_browser_label.setText(_translate("LandingPage", "Login using a browser."))
        self.login_import_radio.setText(_translate("LandingPage", "Import"))
        self.login_import_label.setText(_translate("LandingPage", "Import from Epic Games Launcher"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LandingPage = QtWidgets.QWidget()
    ui = Ui_LandingPage()
    ui.setupUi(LandingPage)
    LandingPage.show()
    sys.exit(app.exec_())



import logging
import os
import queue
import time
from multiprocessing import Queue as MPQueue
from multiprocessing.shared_memory import SharedMemory
from sys import exit
from threading import Condition, Thread

from legendary.downloader.mp.manager import DLManager as DLManagerReal
from legendary.downloader.mp.workers import DLWorker, FileWorker
from legendary.models.downloading import ChunkTask, SharedMemorySegment, TerminateWorkerTask

from rare.lgndr.glue.monkeys import DLManagerSignals
from rare.lgndr.models.downloading import UIUpdate


# fmt: off
class DLManager(DLManagerReal):
    def __init__(self, *args, **kwargs):
        super(DLManager, self).__init__(*args, **kwargs)
        self.log.info("Using Rare's DLManager monkey")

    # Rare: prototype to avoid undefined variable in type checkers
    signals_queue: MPQueue

    # @staticmethod
    def run_real(self):
        self.shared_memory = SharedMemory(create=True, size=self.max_shared_memory)
        self.log.debug(f'Created shared memory of size: {self.shared_memory.size / 1024 / 1024:.02f} MiB')

        # create the shared memory segments and add them to their respective pools
        for i in range(int(self.shared_memory.size / self.analysis.biggest_chunk)):
            _sms = SharedMemorySegment(offset=i * self.analysis.biggest_chunk,
                                       end=i * self.analysis.biggest_chunk + self.analysis.biggest_chunk)
            self.sms.append(_sms)

        self.log.debug(f'Created {len(self.sms)} shared memory segments.')

        # Create queues
        self.dl_worker_queue = MPQueue(-1)
        self.writer_queue = MPQueue(-1)
        self.dl_result_q = MPQueue(-1)
        self.writer_result_q = MPQueue(-1)

        self.log.info(f'Starting download workers...')

        bind_ip = None
        for i in range(self.max_workers):
            if self.bind_ips:
                bind_ip = self.bind_ips[i % len(self.bind_ips)]

            w = DLWorker(f'DLWorker {i + 1}', self.dl_worker_queue, self.dl_result_q,
                         self.shared_memory.name, logging_queue=self.logging_queue,
                         dl_timeout=self.dl_timeout, bind_addr=bind_ip)
            self.children.append(w)
            w.start()

        self.log.info('Starting file writing worker...')
        writer_p = FileWorker(self.writer_queue, self.writer_result_q, self.dl_dir,
                              self.shared_memory.name, self.cache_dir, self.logging_queue)
        self.children.append(writer_p)
        writer_p.start()

        num_chunk_tasks = sum(isinstance(t, ChunkTask) for t in self.tasks)
        num_dl_tasks = len(self.chunks_to_dl)
        num_tasks = len(self.tasks)
        num_shared_memory_segments = len(self.sms)
        self.log.debug(f'Chunks to download: {num_dl_tasks}, File tasks: {num_tasks}, Chunk tasks: {num_chunk_tasks}')

        # active downloader tasks
        self.active_tasks = 0
        processed_chunks = 0
        processed_tasks = 0
        total_dl = 0
        total_write = 0

        # synchronization conditions
        shm_cond = Condition()
        task_cond = Condition()
        self.conditions = [shm_cond, task_cond]

        # start threads
        s_time = time.perf_counter()
        self.threads.append(Thread(target=self.download_job_manager, args=(task_cond, shm_cond)))
        self.threads.append(Thread(target=self.dl_results_handler, args=(task_cond,)))
        self.threads.append(Thread(target=self.fw_results_handler, args=(shm_cond,)))

        for t in self.threads:
            t.start()

        last_update = time.perf_counter()

        # Rare: kill requested
        kill_request = False

        while processed_tasks < num_tasks:
            delta = time.perf_counter() - last_update
            if not delta:
                time.sleep(self.update_interval)
                continue

            # update all the things
            processed_chunks += self.num_processed_since_last
            processed_tasks += self.num_tasks_processed_since_last

            total_dl += self.bytes_downloaded_since_last
            total_write += self.bytes_written_since_last

            dl_speed = self.bytes_downloaded_since_last / delta
            dl_unc_speed = self.bytes_decompressed_since_last / delta
            w_speed = self.bytes_written_since_last / delta
            r_speed = self.bytes_read_since_last / delta
            # c_speed = self.num_processed_since_last / delta

            # set temporary counters to 0
            self.bytes_read_since_last = self.bytes_written_since_last = 0
            self.bytes_downloaded_since_last = self.num_processed_since_last = 0
            self.bytes_decompressed_since_last = self.num_tasks_processed_since_last = 0
            last_update = time.perf_counter()

            perc = (processed_chunks / num_chunk_tasks) * 100
            runtime = time.perf_counter() - s_time
            total_avail = len(self.sms)
            total_used = (num_shared_memory_segments - total_avail) * (self.analysis.biggest_chunk / 1024 / 1024)

            if runtime and processed_chunks:
                average_speed = processed_chunks / runtime
                estimate = (num_chunk_tasks - processed_chunks) / average_speed
                hours, estimate = int(estimate // 3600), estimate % 3600
                minutes, seconds = int(estimate // 60), int(estimate % 60)

                rt_hours, runtime = int(runtime // 3600), runtime % 3600
                rt_minutes, rt_seconds = int(runtime // 60), int(runtime % 60)
            else:
                estimate = 0
                hours = minutes = seconds = 0
                rt_hours = rt_minutes = rt_seconds = 0

            # Rare: Disable up to INFO logging level for the segment below
            log_level = self.log.level
            self.log.setLevel(logging.ERROR)
            self.log.info(f'= Progress: {perc:.02f}% ({processed_chunks}/{num_chunk_tasks}), '
                          f'Running for {rt_hours:02d}:{rt_minutes:02d}:{rt_seconds:02d}, '
                          f'ETA: {hours:02d}:{minutes:02d}:{seconds:02d}')
            self.log.info(f' - Downloaded: {total_dl / 1024 / 1024:.02f} MiB, '
                          f'Written: {total_write / 1024 / 1024:.02f} MiB')
            self.log.info(f' - Cache usage: {total_used:.02f} MiB, active tasks: {self.active_tasks}')
            self.log.info(f' + Download\t- {dl_speed / 1024 / 1024:.02f} MiB/s (raw) '
                          f'/ {dl_unc_speed / 1024 / 1024:.02f} MiB/s (decompressed)')
            self.log.info(f' + Disk\t- {w_speed / 1024 / 1024:.02f} MiB/s (write) / '
                          f'{r_speed / 1024 / 1024:.02f} MiB/s (read)')
            # Rare: Restore previous logging level
            self.log.setLevel(log_level)

            # send status update to back to instantiator (if queue exists)
            if self.status_queue:
                try:
                    self.status_queue.put(UIUpdate(
                        progress=perc, download_speed=dl_unc_speed, write_speed=w_speed, read_speed=r_speed,
                        runtime=round(runtime),
                        estimated_time_left=round(estimate),
                        processed_chunks=processed_chunks,
                        chunk_tasks=num_chunk_tasks,
                        total_downloaded=total_dl,
                        total_written=total_write,
                        cache_usage=total_used,
                        active_tasks=self.active_tasks,
                        download_compressed_speed=dl_speed,
                        memory_usage=total_used * 1024 * 1024
                    ), timeout=1.0)
                except Exception as e:
                    self.log.warning(f'Failed to send status update to queue: {e!r}')

            # Rare: queue of control signals
            try:
                signals: DLManagerSignals = self.signals_queue.get(timeout=0.5)
                self.log.warning('Immediate stop requested!')
                if signals.kill is True:
                    # lk: graceful but not what legendary does
                    self.running = False
                    # send conditions to unlock threads if they aren't already
                    for cond in self.conditions:
                        with cond:
                            cond.notify()
                    kill_request = True
                    break
                    # # lk: alternative way, but doesn't clean shm
                    # for i in range(self.max_workers):
                    #     self.dl_worker_queue.put_nowait(TerminateWorkerTask())
                    #
                    # self.log.info('Waiting for installation to finish...')
                    # self.writer_queue.put_nowait(TerminateWorkerTask())
                    # raise KeyboardInterrupt
            except queue.Empty:
                pass

            time.sleep(self.update_interval)

        for i in range(self.max_workers):
            self.dl_worker_queue.put_nowait(TerminateWorkerTask())

        self.log.info('Waiting for installation to finish...')
        self.writer_queue.put_nowait(TerminateWorkerTask())

        writer_p.join(timeout=10.0)
        if writer_p.exitcode is None:
            self.log.warning(f'Terminating writer process, no exit code!')
            writer_p.terminate()

        # forcibly kill DL workers that are not actually dead yet
        for child in self.children:
            if child.exitcode is None:
                child.terminate()

        # make sure all the threads are dead.
        for t in self.threads:
            t.join(timeout=5.0)
            if t.is_alive():
                self.log.warning(f'Thread did not terminate! {repr(t)}')

        # clean up resume file
        if self.resume_file and not kill_request:
            try:
                os.remove(self.resume_file)
            except OSError as e:
                self.log.warning(f'Failed to remove resume file: {e!r}')

        # close up shared memory
        self.shared_memory.close()
        self.shared_memory.unlink()
        self.shared_memory = None

        self.log.info('All done! Download manager quitting...')
        # finally, exit the process.
        exit(0)

# fmt: on



import os
import platform
from logging import getLogger
from typing import List

from PyQt5.QtCore import QRunnable, QObject, pyqtSignal, QThreadPool
from PyQt5.QtWidgets import QGroupBox, QMessageBox
from legendary.lfs import eos

from rare.models.install import InstallOptionsModel
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton
from rare.ui.components.tabs.games.integrations.eos_widget import Ui_EosWidget

logger = getLogger("EOS")


def get_wine_prefixes() -> List[str]:
    prefixes = list()
    if os.path.exists(p := os.path.expanduser("~/.wine")):
        prefixes.append(p)

    for name, section in LegendaryCoreSingleton().lgd.config.items():
        pfx = section.get("WINEPREFIX") or section.get("wine_prefix")
        if pfx and pfx not in prefixes:
            prefixes.append(pfx)

    return prefixes


class CheckForUpdateWorker(QRunnable):
    class CheckForUpdateSignals(QObject):
        update_available = pyqtSignal(bool)

    def __init__(self):
        super(CheckForUpdateWorker, self).__init__()
        self.signals = self.CheckForUpdateSignals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()

    def run(self) -> None:
        self.core.check_for_overlay_updates()
        self.signals.update_available.emit(self.core.overlay_update_available)


class EOSGroup(QGroupBox, Ui_EosWidget):
    def __init__(self, parent=None):
        super(EOSGroup, self).__init__(parent=parent)
        self.setupUi(self)
        # lk: set object names for CSS properties
        self.install_button.setObjectName("InstallButton")
        self.uninstall_button.setObjectName("UninstallButton")

        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.prefix_enabled = False

        self.enabled_cb.stateChanged.connect(self.change_enable)
        self.uninstall_button.clicked.connect(self.uninstall_overlay)

        self.update_button.setVisible(False)
        self.overlay = self.core.lgd.get_overlay_install_info()

        self.signals.application.overlay_installed.connect(self.overlay_installation_finished)
        self.signals.application.prefix_updated.connect(self.update_prefixes)

        self.update_check_button.clicked.connect(self.check_for_update)
        self.install_button.clicked.connect(self.install_overlay)
        self.update_button.clicked.connect(lambda: self.install_overlay(True))

        if self.overlay:  # installed
            self.installed_version_lbl.setText(f"<b>{self.overlay.version}</b>")
            self.installed_path_lbl.setText(f"<b>{self.overlay.install_path}</b>")
            self.overlay_stack.setCurrentIndex(0)
        else:
            self.overlay_stack.setCurrentIndex(1)
            self.enable_frame.setDisabled(True)

        if platform.system() == "Windows":
            self.current_prefix = None
            self.select_pfx_combo.setVisible(False)
        else:
            self.current_prefix = os.path.expanduser("~/.wine") \
                if os.path.exists(os.path.expanduser("~/.wine")) \
                else None
            pfxs = get_wine_prefixes()
            for pfx in pfxs:
                self.select_pfx_combo.addItem(pfx.replace(os.path.expanduser("~/"), "~/"))
            if not pfxs:
                self.enable_frame.setDisabled(True)
            else:
                self.select_pfx_combo.setCurrentIndex(0)

            self.select_pfx_combo.currentIndexChanged.connect(self.update_select_combo)
            if pfxs:
                self.update_select_combo(None)

        self.enabled_info_label.setText("")

        self.threadpool = QThreadPool.globalInstance()

    def update_prefixes(self):
        logger.debug("Updated prefixes")
        pfxs = get_wine_prefixes()  # returns /home/whatever
        self.select_pfx_combo.clear()

        for pfx in pfxs:
            self.select_pfx_combo.addItem(pfx.replace(os.path.expanduser("~/"), "~/"))

        if self.current_prefix in pfxs:
            self.select_pfx_combo.setCurrentIndex(
                self.select_pfx_combo.findText(self.current_prefix.replace(os.path.expanduser("~/"), "~/")))

    def check_for_update(self):
        def worker_finished(update_available):
            self.update_button.setVisible(update_available)
            self.update_check_button.setDisabled(False)
            if not update_available:
                self.update_check_button.setText(self.tr("No update available"))

        self.update_check_button.setDisabled(True)
        worker = CheckForUpdateWorker()
        worker.signals.update_available.connect(worker_finished)
        QThreadPool.globalInstance().start(worker)

    def overlay_installation_finished(self):
        self.overlay = self.core.lgd.get_overlay_install_info()

        if not self.overlay:
            logger.error("Something went wrong, when installing overlay")
            QMessageBox.warning(self, "Error", self.tr("Something went wrong, when installing overlay"))
            return

        self.overlay_stack.setCurrentIndex(0)
        self.installed_version_lbl.setText(f"<b>{self.overlay.version}</b>")
        self.installed_path_lbl.setText(f"<b>{self.overlay.install_path}</b>")

        self.update_button.setVisible(False)

        self.enable_frame.setEnabled(True)

    def update_select_combo(self, i: None):
        if i is None:
            i = self.select_pfx_combo.currentIndex()
        prefix = os.path.expanduser(self.select_pfx_combo.itemText(i))
        if platform.system() != "Windows" and not os.path.isfile(os.path.join(prefix, "user.reg")):
            return
        self.current_prefix = prefix
        reg_paths = eos.query_registry_entries(self.current_prefix)

        overlay_enabled = False
        if reg_paths['overlay_path'] and self.core.is_overlay_install(reg_paths['overlay_path']):
            overlay_enabled = True
        self.enabled_cb.setChecked(overlay_enabled)

    def change_enable(self):
        enabled = self.enabled_cb.isChecked()
        if not enabled:
            try:
                eos.remove_registry_entries(self.current_prefix)
            except PermissionError:
                logger.error("Can't disable eos overlay")
                QMessageBox.warning(self, "Error", self.tr(
                    "Failed to disable Overlay. Probably it is installed by Epic Games Launcher"))
                return
            logger.info("Disabled Epic Overlay")
            self.enabled_info_label.setText(self.tr("Disabled"))
        else:
            if not self.overlay:
                available_installs = self.core.search_overlay_installs(self.current_prefix)
                if not available_installs:
                    logger.error('No EOS overlay installs found!')
                    return
                path = available_installs[0]
            else:
                path = self.overlay.install_path

            if not self.core.is_overlay_install(path):
                logger.error(f'Not a valid Overlay installation: {path}')
                self.select_pfx_combo.removeItem(self.select_pfx_combo.currentIndex())
                return

            path = os.path.normpath(path)
            reg_paths = eos.query_registry_entries(self.current_prefix)
            if old_path := reg_paths["overlay_path"]:
                if os.path.normpath(old_path) == path:
                    logger.info(f'Overlay already enabled, nothing to do.')
                    return
                else:
                    logger.info(f'Updating overlay registry entries from "{old_path}" to "{path}"')
                try:
                    eos.remove_registry_entries(self.current_prefix)
                except PermissionError:
                    logger.error("Can't disable eos overlay")
                    QMessageBox.warning(self, "Error", self.tr(
                        "Failed to disable Overlay. Probably it is installed by Epic Games Launcher"))
                    return
            try:
                eos.add_registry_entries(path, self.current_prefix)
            except PermissionError:
                logger.error("Failed to disable eos overlay")
                QMessageBox.warning(self, "Error", self.tr(
                    "Failed to enable EOS overlay. Maybe it is already installed by Epic Games Launcher"))
                return
            self.enabled_info_label.setText(self.tr("Enabled"))
            logger.info(f'Enabled overlay at: {path}')

    def update_checkbox(self):
        reg_paths = eos.query_registry_entries(self.current_prefix)
        enabled = False
        if reg_paths['overlay_path'] and self.core.is_overlay_install(reg_paths['overlay_path']):
            enabled = True
        self.enabled_cb.setChecked(enabled)

    def install_overlay(self, update=False):
        base_path = os.path.join(self.core.get_default_install_dir(), ".overlay")
        if update:
            if not self.overlay:
                self.overlay_stack.setCurrentIndex(1)
                self.enable_frame.setDisabled(True)
                QMessageBox.warning(self, "Warning", self.tr("Overlay is not installed. Could not update"))
                return
            base_path = self.overlay.install_path

        options = InstallOptionsModel(
            app_name=eos.EOSOverlayApp.app_name, base_path=base_path, platform="Windows", overlay=True
        )

        self.signals.game.install.emit(options)

    def uninstall_overlay(self):
        if not self.core.is_overlay_installed():
            logger.error('No legendary-managed overlay installation found.')
            self.overlay_stack.setCurrentIndex(1)
            return

        if QMessageBox.No == QMessageBox.question(
                self, "Uninstall Overlay", self.tr("Do you want to uninstall overlay?"),
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No
        ):
            return
        if platform.system() == "Windows":
            eos.remove_registry_entries(None)
        else:
            for prefix in [self.select_pfx_combo.itemText(i) for i in range(self.select_pfx_combo.count())]:
                logger.info(f"Removing registry entries from {prefix}")
                try:
                    eos.remove_registry_entries(os.path.expanduser(prefix))
                except Exception as e:
                    logger.warning(f"{prefix}: {e}")

        self.core.remove_overlay_install()
        self.overlay_stack.setCurrentIndex(1)

        self.enable_frame.setDisabled(True)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/game_info/sync_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SyncWidget(object):
    def setupUi(self, SyncWidget):
        SyncWidget.setObjectName("SyncWidget")
        SyncWidget.resize(438, 137)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SyncWidget.sizePolicy().hasHeightForWidth())
        SyncWidget.setSizePolicy(sizePolicy)
        SyncWidget.setWindowTitle("SyncWidget")
        self.sync_layout = QtWidgets.QHBoxLayout(SyncWidget)
        self.sync_layout.setContentsMargins(0, 0, 0, 0)
        self.sync_layout.setObjectName("sync_layout")
        self.local_gb = QtWidgets.QGroupBox(SyncWidget)
        self.local_gb.setObjectName("local_gb")
        self.local_layout = QtWidgets.QVBoxLayout(self.local_gb)
        self.local_layout.setObjectName("local_layout")
        self.date_info_local = QtWidgets.QLabel(self.local_gb)
        self.date_info_local.setText("")
        self.date_info_local.setAlignment(QtCore.Qt.AlignCenter)
        self.date_info_local.setObjectName("date_info_local")
        self.local_layout.addWidget(self.date_info_local)
        self.icon_local = QtWidgets.QLabel(self.local_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.icon_local.sizePolicy().hasHeightForWidth())
        self.icon_local.setSizePolicy(sizePolicy)
        self.icon_local.setText("")
        self.icon_local.setAlignment(QtCore.Qt.AlignCenter)
        self.icon_local.setObjectName("icon_local")
        self.local_layout.addWidget(self.icon_local)
        self.age_label_local = QtWidgets.QLabel(self.local_gb)
        self.age_label_local.setText("")
        self.age_label_local.setAlignment(QtCore.Qt.AlignCenter)
        self.age_label_local.setObjectName("age_label_local")
        self.local_layout.addWidget(self.age_label_local)
        self.upload_button = QtWidgets.QPushButton(self.local_gb)
        self.upload_button.setMinimumSize(QtCore.QSize(192, 0))
        self.upload_button.setObjectName("upload_button")
        self.local_layout.addWidget(self.upload_button)
        self.sync_layout.addWidget(self.local_gb)
        self.cloud_gb = QtWidgets.QGroupBox(SyncWidget)
        self.cloud_gb.setObjectName("cloud_gb")
        self.cloud_layout = QtWidgets.QVBoxLayout(self.cloud_gb)
        self.cloud_layout.setObjectName("cloud_layout")
        self.date_info_remote = QtWidgets.QLabel(self.cloud_gb)
        self.date_info_remote.setText("")
        self.date_info_remote.setAlignment(QtCore.Qt.AlignCenter)
        self.date_info_remote.setObjectName("date_info_remote")
        self.cloud_layout.addWidget(self.date_info_remote)
        self.icon_remote = QtWidgets.QLabel(self.cloud_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.icon_remote.sizePolicy().hasHeightForWidth())
        self.icon_remote.setSizePolicy(sizePolicy)
        self.icon_remote.setText("")
        self.icon_remote.setAlignment(QtCore.Qt.AlignCenter)
        self.icon_remote.setObjectName("icon_remote")
        self.cloud_layout.addWidget(self.icon_remote)
        self.age_label_remote = QtWidgets.QLabel(self.cloud_gb)
        self.age_label_remote.setText("")
        self.age_label_remote.setAlignment(QtCore.Qt.AlignCenter)
        self.age_label_remote.setObjectName("age_label_remote")
        self.cloud_layout.addWidget(self.age_label_remote)
        self.download_button = QtWidgets.QPushButton(self.cloud_gb)
        self.download_button.setMinimumSize(QtCore.QSize(192, 0))
        self.download_button.setObjectName("download_button")
        self.cloud_layout.addWidget(self.download_button)
        self.sync_layout.addWidget(self.cloud_gb)

        self.retranslateUi(SyncWidget)

    def retranslateUi(self, SyncWidget):
        _translate = QtCore.QCoreApplication.translate
        self.local_gb.setTitle(_translate("SyncWidget", "Local"))
        self.upload_button.setText(_translate("SyncWidget", "Upload"))
        self.cloud_gb.setTitle(_translate("SyncWidget", "Cloud"))
        self.download_button.setText(_translate("SyncWidget", "Download"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    SyncWidget = QtWidgets.QWidget()
    ui = Ui_SyncWidget()
    ui.setupUi(SyncWidget)
    SyncWidget.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/launch_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LaunchDialog(object):
    def setupUi(self, LaunchDialog):
        LaunchDialog.setObjectName("LaunchDialog")
        LaunchDialog.resize(400, 160)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LaunchDialog.sizePolicy().hasHeightForWidth())
        LaunchDialog.setSizePolicy(sizePolicy)
        LaunchDialog.setMinimumSize(QtCore.QSize(400, 160))
        LaunchDialog.setMaximumSize(QtCore.QSize(400, 160))
        self.launch_layout = QtWidgets.QVBoxLayout(LaunchDialog)
        self.launch_layout.setObjectName("launch_layout")
        self.title_label = QtWidgets.QLabel(LaunchDialog)
        self.title_label.setObjectName("title_label")
        self.launch_layout.addWidget(self.title_label)
        self.progress_bar = QtWidgets.QProgressBar(LaunchDialog)
        self.progress_bar.setProperty("value", 0)
        self.progress_bar.setObjectName("progress_bar")
        self.launch_layout.addWidget(self.progress_bar)

        self.retranslateUi(LaunchDialog)

    def retranslateUi(self, LaunchDialog):
        _translate = QtCore.QCoreApplication.translate
        LaunchDialog.setWindowTitle(_translate("LaunchDialog", "Launching - Rare"))
        self.title_label.setText(_translate("LaunchDialog", "<h2>Launching Rare</h2>"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    LaunchDialog = QtWidgets.QDialog()
    ui = Ui_LaunchDialog()
    ui.setupUi(LaunchDialog)
    LaunchDialog.show()
    sys.exit(app.exec_())



from dataclasses import dataclass
from typing import Optional


@dataclass
class UIUpdate:
    """
    Status update object sent from the manager to the CLI/GUI to update status indicators
    Inheritance doesn't work due to optional arguments in UIUpdate proper
    """
    progress: float
    download_speed: float
    write_speed: float
    read_speed: float
    memory_usage: float
    runtime: float
    estimated_time_left: float
    processed_chunks: int
    chunk_tasks: int
    total_downloaded: float
    total_written: float
    cache_usage: float
    active_tasks: int
    download_compressed_speed: float
    current_filename: Optional[str] = None



from typing import Callable, Optional

from legendary.core import LegendaryCore
from legendary.models.config import LGDConf

_config: Optional[LGDConf] = None
_save_config: Optional[Callable[[], None]] = None


def init_config_handler(core: LegendaryCore):
    global _config, _save_config
    _config = core.lgd.config
    _save_config = core.lgd.save_config


def save_config():
    if _save_config is None:
        raise RuntimeError("Uninitialized use of config_helper")
    _save_config()


def add_option(app_name: str, option: str, value: str):
    value = value.replace("%%", "%").replace("%", "%%")
    if not _config.has_section(app_name):
        _config[app_name] = {}

    _config.set(app_name, option, value)
    save_config()


def remove_option(app_name, option):
    if _config.has_option(app_name, option):
        _config.remove_option(app_name, option)

    # if _config.has_section(app_name) and not _config[app_name]:
    #     _config.remove_section(app_name)

    save_config()


def remove_section(app_name):
    return
    if _config.has_section(app_name):
        _config.remove_section(app_name)
        save_config()



from logging import getLogger

from PyQt5.QtCore import Qt, QEvent, QRect
from PyQt5.QtGui import (
    QPalette,
    QBrush,
    QPaintEvent,
    QPainter,
    QLinearGradient,
    QPixmap,
    QImage,
    QResizeEvent,
)

from rare.models.game import RareGame
from rare.utils.misc import format_size
from rare.widgets.image_widget import ImageWidget
from .game_widget import GameWidget
from .list_widget import ListWidget

logger = getLogger("ListGameWidget")


class ListGameWidget(GameWidget):
    def __init__(self, rgame: RareGame, parent=None):
        super().__init__(rgame, parent)
        self.setObjectName(f"{rgame.app_name}")
        self.ui = ListWidget()
        self.ui.setupUi(self)

        self.ui.title_label.setText(self.rgame.app_title)
        self.ui.launch_btn.clicked.connect(self._launch)
        self.ui.launch_btn.setVisible(self.rgame.is_installed)
        self.ui.install_btn.clicked.connect(self._install)
        self.ui.install_btn.setVisible(not self.rgame.is_installed)

        self.ui.launch_btn.setEnabled(self.rgame.can_launch)

        self.ui.launch_btn.setText(
            self.tr("Launch") if not self.rgame.is_origin else self.tr("Link/Play")
        )
        self.ui.developer_label.setText(self.rgame.developer)
        # self.version_label.setVisible(self.is_installed)
        if self.rgame.igame:
            self.ui.version_label.setText(self.rgame.version)
        self.ui.size_label.setText(format_size(self.rgame.install_size) if self.rgame.install_size else "")

        self.update_state()

        # lk: "connect" the buttons' enter/leave events to this widget
        self.installEventFilter(self)
        self.ui.launch_btn.installEventFilter(self)
        self.ui.install_btn.installEventFilter(self)

    def enterEvent(self, a0: QEvent = None) -> None:
        if a0 is not None:
            a0.accept()
        self.ui.tooltip_label.setVisible(True)

    def leaveEvent(self, a0: QEvent = None) -> None:
        if a0 is not None:
            a0.accept()
        self.ui.tooltip_label.setVisible(False)

    """
    Painting and progress overrides.
    Let them live here until a better alternative is divised.
    
    The list widget and these painting functions can be
    refactored to be used in downloads and/or dlcs
    """

    def event(self, e: QEvent) -> bool:
        if e.type() == QEvent.LayoutRequest:
            if self.progress_label.isVisible():
                width = int(self._pixmap.width() / self._pixmap.devicePixelRatioF())
                origin = self.width() - width
                fill_rect = QRect(origin, 0, width, self.sizeHint().height())
                self.progress_label.setGeometry(fill_rect)
        return ImageWidget.event(self, e)

    def resizeEvent(self, a0: QResizeEvent) -> None:
        if self.progress_label.isVisible():
            width = int(self._pixmap.width() / self._pixmap.devicePixelRatioF())
            origin = self.width() - width
            fill_rect = QRect(origin, 0, width, self.sizeHint().height())
            self.progress_label.setGeometry(fill_rect)
        ImageWidget.resizeEvent(self, a0)

    def prepare_pixmap(self, pixmap: QPixmap) -> QPixmap:
        device: QImage = QImage(
            pixmap.size().width() * 3,
            int(self.sizeHint().height() * pixmap.devicePixelRatioF()) + 1,
            QImage.Format_ARGB32_Premultiplied
        )
        painter = QPainter(device)
        brush = QBrush(pixmap)
        painter.fillRect(device.rect(), brush)
        # the gradient could be cached and reused as it is expensive
        gradient = QLinearGradient(0, 0, device.width(), 0)
        gradient.setColorAt(0.15, Qt.transparent)
        gradient.setColorAt(0.5, Qt.black)
        gradient.setColorAt(0.85, Qt.transparent)
        painter.setCompositionMode(QPainter.CompositionMode_DestinationIn)
        painter.fillRect(device.rect(), gradient)
        painter.end()
        ret = QPixmap.fromImage(device)
        ret.setDevicePixelRatio(pixmap.devicePixelRatioF())
        return ret

    def setPixmap(self, pixmap: QPixmap) -> None:
        # lk: trade some possible delay and start-up time
        # lk: for faster rendering. Gradients are expensive
        # lk: so pre-generate the image
        super(ListGameWidget, self).setPixmap(self.prepare_pixmap(pixmap))

    def paint_image_cover(self, painter: QPainter, a0: QPaintEvent) -> None:
        painter.setOpacity(self._opacity)
        color = self.palette().color(QPalette.Background).darker(75)
        painter.fillRect(self.rect(), color)
        brush = QBrush(self._pixmap)
        brush.setTransform(self._transform)
        width = int(self._pixmap.width() / self._pixmap.devicePixelRatioF())
        origin = self.width() - width
        painter.setBrushOrigin(origin, 0)
        fill_rect = QRect(origin, 0, width, self.height())
        painter.fillRect(fill_rect, brush)

    def progressPixmap(self, color: QPixmap, gray: QPixmap, progress: int) -> QPixmap:
        # lk: so about that +1 after the in convertion, casting to int rounds down
        # lk: and that can create a weird line at the bottom, add 1 to round up.
        device = QPixmap(
            color.size().width(),
            int(self.sizeHint().height() * color.devicePixelRatioF()) + 1,
        )
        painter = QPainter(device)
        painter.setRenderHint(QPainter.SmoothPixmapTransform, self._smooth_transform)
        painter.setCompositionMode(QPainter.CompositionMode_Source)
        prog_h = (device.height() * progress // 100)
        brush = QBrush(gray)
        painter.fillRect(device.rect().adjusted(0, 0, 0, -prog_h), brush)
        brush.setTexture(color)
        painter.fillRect(device.rect().adjusted(0, device.height() - prog_h, 0, 0), brush)
        painter.end()
        device.setDevicePixelRatio(color.devicePixelRatioF())
        return device



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/store/wishlist.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Wishlist(object):
    def setupUi(self, Wishlist):
        Wishlist.setObjectName("Wishlist")
        Wishlist.resize(736, 398)
        Wishlist.setWindowTitle("StackedWidget")
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scroll_area = QtWidgets.QScrollArea(self.page)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setObjectName("scroll_area")
        self.scroll_widget = QtWidgets.QWidget()
        self.scroll_widget.setGeometry(QtCore.QRect(0, 0, 716, 378))
        self.scroll_widget.setObjectName("scroll_widget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scroll_widget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.title_label = QtWidgets.QLabel(self.scroll_widget)
        font = QtGui.QFont()
        font.setPointSize(15)
        self.title_label.setFont(font)
        self.title_label.setObjectName("title_label")
        self.verticalLayout_2.addWidget(self.title_label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.sort_label = QtWidgets.QLabel(self.scroll_widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sort_label.sizePolicy().hasHeightForWidth())
        self.sort_label.setSizePolicy(sizePolicy)
        self.sort_label.setObjectName("sort_label")
        self.horizontalLayout.addWidget(self.sort_label)
        self.sort_cb = QtWidgets.QComboBox(self.scroll_widget)
        self.sort_cb.setObjectName("sort_cb")
        self.sort_cb.addItem("")
        self.sort_cb.addItem("")
        self.sort_cb.addItem("")
        self.sort_cb.addItem("")
        self.horizontalLayout.addWidget(self.sort_cb)
        self.reverse = QtWidgets.QCheckBox(self.scroll_widget)
        self.reverse.setObjectName("reverse")
        self.horizontalLayout.addWidget(self.reverse)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.filter_label = QtWidgets.QLabel(self.scroll_widget)
        self.filter_label.setObjectName("filter_label")
        self.horizontalLayout.addWidget(self.filter_label)
        self.filter_cb = QtWidgets.QComboBox(self.scroll_widget)
        self.filter_cb.setObjectName("filter_cb")
        self.filter_cb.addItem("")
        self.filter_cb.addItem("")
        self.horizontalLayout.addWidget(self.filter_cb)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.reload_button = QtWidgets.QPushButton(self.scroll_widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.reload_button.sizePolicy().hasHeightForWidth())
        self.reload_button.setSizePolicy(sizePolicy)
        self.reload_button.setText("")
        self.reload_button.setObjectName("reload_button")
        self.horizontalLayout.addWidget(self.reload_button)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.list_layout = QtWidgets.QVBoxLayout()
        self.list_layout.setObjectName("list_layout")
        self.verticalLayout_2.addLayout(self.list_layout)
        self.no_games_label = QtWidgets.QLabel(self.scroll_widget)
        self.no_games_label.setObjectName("no_games_label")
        self.verticalLayout_2.addWidget(self.no_games_label)
        spacerItem2 = QtWidgets.QSpacerItem(379, 218, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem2)
        self.scroll_area.setWidget(self.scroll_widget)
        self.verticalLayout.addWidget(self.scroll_area)
        Wishlist.addWidget(self.page)

        self.retranslateUi(Wishlist)

    def retranslateUi(self, Wishlist):
        _translate = QtCore.QCoreApplication.translate
        self.title_label.setText(_translate("Wishlist", "Wishlist"))
        self.sort_label.setText(_translate("Wishlist", "Sort by"))
        self.sort_cb.setItemText(0, _translate("Wishlist", "Name"))
        self.sort_cb.setItemText(1, _translate("Wishlist", "Price"))
        self.sort_cb.setItemText(2, _translate("Wishlist", "Developer"))
        self.sort_cb.setItemText(3, _translate("Wishlist", "Discount"))
        self.reverse.setText(_translate("Wishlist", "Reverse"))
        self.filter_label.setText(_translate("Wishlist", "Filter:"))
        self.filter_cb.setItemText(0, _translate("Wishlist", "None"))
        self.filter_cb.setItemText(1, _translate("Wishlist", "Discount"))
        self.no_games_label.setText(_translate("Wishlist", "No games matching your filter"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Wishlist = QtWidgets.QStackedWidget()
    ui = Ui_Wishlist()
    ui.setupUi(Wishlist)
    Wishlist.show()
    sys.exit(app.exec_())



from .fetch import FetchWorker
from .install_info import InstallInfoWorker
from .move import MoveWorker
from .uninstall import UninstallWorker
from .verify import VerifyWorker
from .wine_resolver import OriginWineWorker
from .worker import Worker, QueueWorker



import os
import platform
import shutil
from logging import getLogger
from typing import Optional

from PyQt5.QtCore import (
    Qt,
    pyqtSlot,
    pyqtSignal,
)
from PyQt5.QtWidgets import (
    QMenu,
    QWidget,
    QMessageBox,
    QWidgetAction,
)

from rare.models.game import RareGame
from rare.shared import RareCore
from rare.shared.workers import VerifyWorker, MoveWorker
from rare.ui.components.tabs.games.game_info.game_info import Ui_GameInfo
from rare.utils.misc import format_size, icon
from rare.widgets.image_widget import ImageWidget, ImageSize
from rare.widgets.side_tab import SideTabContents
from .move_game import MoveGamePopUp, is_game_dir

logger = getLogger("GameInfo")


class GameInfo(QWidget, SideTabContents):
    # str: app_name
    import_clicked = pyqtSignal(str)

    def __init__(self, parent=None):
        super(GameInfo, self).__init__(parent=parent)
        self.ui = Ui_GameInfo()
        self.ui.setupUi(self)
        # lk: set object names for CSS properties
        self.ui.install_button.setObjectName("InstallButton")
        self.ui.modify_button.setObjectName("InstallButton")
        self.ui.uninstall_button.setObjectName("UninstallButton")

        self.ui.install_button.setIcon(icon("ri.install-line"))
        self.ui.import_button.setIcon(icon("mdi.file-import"))

        self.ui.modify_button.setIcon(icon("fa.gear"))
        self.ui.verify_button.setIcon(icon("fa.check"))
        self.ui.repair_button.setIcon(icon("fa.wrench"))
        self.ui.move_button.setIcon(icon("mdi.folder-move"))
        self.ui.uninstall_button.setIcon(icon("ri.uninstall-line"))

        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.args = RareCore.instance().args()
        # self.image_manager = RareCore.instance().image_manager()

        self.rgame: Optional[RareGame] = None

        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Display)
        self.ui.left_layout.insertWidget(0, self.image, alignment=Qt.AlignTop)

        self.ui.install_button.clicked.connect(self.__on_install)
        self.ui.import_button.clicked.connect(self.__on_import)
        self.ui.modify_button.clicked.connect(self.__on_modify)
        self.ui.verify_button.clicked.connect(self.__on_verify)
        self.ui.repair_button.clicked.connect(self.__on_repair)
        self.ui.uninstall_button.clicked.connect(self.__on_uninstall)

        self.move_game_pop_up = MoveGamePopUp(self)
        move_action = QWidgetAction(self)
        move_action.setDefaultWidget(self.move_game_pop_up)
        self.ui.move_button.setMenu(QMenu(self.ui.move_button))
        self.ui.move_button.menu().addAction(move_action)

        self.move_game_pop_up.browse_done.connect(self.ui.move_button.showMenu)
        self.move_game_pop_up.move_clicked.connect(self.ui.move_button.menu().close)
        self.move_game_pop_up.move_clicked.connect(self.__on_move)

        self.steam_grade_ratings = {
            "platinum": self.tr("Platinum"),
            "gold": self.tr("Gold"),
            "silver": self.tr("Silver"),
            "bronze": self.tr("Bronze"),
            "borked": self.tr("Borked"),
            "fail": self.tr("Failed to get rating"),
            "pending": self.tr("Loading..."),
            "na": self.tr("Not applicable"),
        }

        # lk: hide unfinished things
        self.ui.tags_group.setVisible(False)
        self.ui.requirements_group.setVisible(False)

    @pyqtSlot()
    def __on_install(self):
        if self.rgame.is_non_asset:
            self.rgame.launch()
        else:
            self.rgame.install()

    @pyqtSlot()
    def __on_import(self):
        self.import_clicked.emit(self.rgame.app_name)

    @pyqtSlot()
    def __on_uninstall(self):
        """ This method is to be called from the button only """
        self.rgame.uninstall()

    @pyqtSlot()
    def __on_modify(self):
        """ This method is to be called from the button only """
        self.rgame.modify()

    @pyqtSlot()
    def __on_repair(self):
        """ This method is to be called from the button only """
        repair_file = os.path.join(self.core.lgd.get_tmp_path(), f"{self.rgame.app_name}.repair")
        if not os.path.exists(repair_file):
            QMessageBox.warning(
                self,
                self.tr("Error - {}").format(self.rgame.app_title),
                self.tr(
                    "Repair file does not exist or game does not need a repair. Please verify game first"
                ),
            )
            return
        self.repair_game(self.rgame)

    def repair_game(self, rgame: RareGame):
        rgame.update_game()
        ans = False
        if rgame.has_update:
            ans = QMessageBox.question(
                self,
                self.tr("Repair and update? - {}").format(self.rgame.app_title),
                self.tr(
                    "There is an update for <b>{}</b> from <b>{}</b> to <b>{}</b>. "
                    "Do you want to update the game while repairing it?"
                ).format(rgame.app_title, rgame.version, rgame.remote_version),
            ) == QMessageBox.Yes
        rgame.repair(repair_and_update=ans)

    @pyqtSlot(RareGame, str)
    def __on_worker_error(self, rgame: RareGame, message: str):
        QMessageBox.warning(
            self,
            self.tr("Error - {}").format(rgame.app_title),
            message
        )

    @pyqtSlot()
    def __on_verify(self):
        """ This method is to be called from the button only """
        if not os.path.exists(self.rgame.igame.install_path):
            logger.error(f"Installation path {self.rgame.igame.install_path} for {self.rgame.app_title} does not exist")
            QMessageBox.warning(
                self,
                self.tr("Error - {}").format(self.rgame.app_title),
                self.tr("Installation path for <b>{}</b> does not exist. Cannot continue.").format(self.rgame.app_title),
            )
            return
        self.verify_game(self.rgame)

    def verify_game(self, rgame: RareGame):
        worker = VerifyWorker(self.core, self.args, rgame)
        worker.signals.progress.connect(self.__on_verify_progress)
        worker.signals.result.connect(self.__on_verify_result)
        worker.signals.error.connect(self.__on_worker_error)
        self.rcore.enqueue_worker(rgame, worker)

    @pyqtSlot(RareGame, int, int, float, float)
    def __on_verify_progress(self, rgame: RareGame, num, total, percentage, speed):
        # lk: the check is NOT REQUIRED because signals are disconnected but protect against it anyway
        if rgame is not self.rgame:
            return
        self.ui.verify_progress.setValue(num * 100 // total)

    @pyqtSlot(RareGame, bool, int, int)
    def __on_verify_result(self, rgame: RareGame, success, failed, missing):
        self.ui.repair_button.setDisabled(success)
        if success:
            QMessageBox.information(
                self,
                self.tr("Summary - {}").format(rgame.app_title),
                self.tr("<b>{}</b> has been verified successfully. "
                        "No missing or corrupt files found").format(rgame.app_title),
            )
        else:
            ans = QMessageBox.question(
                self,
                self.tr("Summary - {}").format(rgame.app_title),
                self.tr(
                    "<b>{}</b> failed verification, <b>{}</b> file(s) corrupted, <b>{}</b> file(s) are missing. "
                    "Do you want to repair them?"
                ).format(rgame.app_title, failed, missing),
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.Yes,
            )
            if ans == QMessageBox.Yes:
                self.repair_game(rgame)

    @pyqtSlot(str)
    def __on_move(self, dst_path: str):
        """ This method is to be called from the button only """
        new_install_path = os.path.join(dst_path, os.path.basename(self.rgame.install_path))

        dir_exists = False
        if os.path.isdir(new_install_path):
            dir_exists = is_game_dir(self.rgame.install_path, new_install_path)

        if not dir_exists:
            for item in os.listdir(dst_path):
                if os.path.basename(self.rgame.install_path) in os.path.basename(item):
                    ans = QMessageBox.question(
                        self,
                        self.tr("Move game? - {}").format(self.rgame.app_title),
                        self.tr(
                            "Destination <b>{}</b> already exists. "
                            "Are you sure you want to overwrite it?"
                        ).format(new_install_path),
                        QMessageBox.Yes | QMessageBox.No,
                        QMessageBox.Yes,
                    )

                    if ans == QMessageBox.Yes:
                        if os.path.isdir(new_install_path):
                            shutil.rmtree(new_install_path)
                        else:
                            os.remove(new_install_path)
                    else:
                        return

        self.move_game(self.rgame, new_install_path, dir_exists)

    def move_game(self, rgame: RareGame, dst_path, dst_exists):
        worker = MoveWorker(
            self.core, rgame=rgame, dst_path=dst_path, dst_exists=dst_exists
        )
        worker.signals.progress.connect(self.__on_move_progress)
        worker.signals.result.connect(self.__on_move_result)
        worker.signals.error.connect(self.__on_worker_error)
        self.rcore.enqueue_worker(self.rgame, worker)

    @pyqtSlot(RareGame, int, object, object)
    def __on_move_progress(self, rgame: RareGame, progress: int, total_size: int, copied_size: int):
        # lk: the check is NOT REQUIRED because signals are disconnected but protect against it anyway
        if rgame is not self.rgame:
            return
        self.ui.move_progress.setValue(progress)

    @pyqtSlot(RareGame, str)
    def __on_move_result(self, rgame: RareGame, dst_path: str):
        QMessageBox.information(
            self,
            self.tr("Summary - {}").format(rgame.app_title),
            self.tr("<b>{}</b> successfully moved to <b>{}<b>.").format(rgame.app_title, dst_path),
        )

    @pyqtSlot()
    def __update_widget(self):
        """ React to state updates from RareGame """
        # self.image.setPixmap(self.image_manager.get_pixmap(self.rgame.app_name, True))
        self.image.setPixmap(self.rgame.pixmap)

        self.ui.lbl_version.setDisabled(self.rgame.is_non_asset)
        self.ui.version.setDisabled(self.rgame.is_non_asset)
        self.ui.version.setText(
            self.rgame.version if not self.rgame.is_non_asset else "N/A"
        )

        self.ui.lbl_install_size.setEnabled(bool(self.rgame.install_size))
        self.ui.install_size.setEnabled(bool(self.rgame.install_size))
        self.ui.install_size.setText(
            format_size(self.rgame.install_size) if self.rgame.install_size else "N/A"
        )

        self.ui.lbl_install_path.setEnabled(bool(self.rgame.install_path))
        self.ui.install_path.setEnabled(bool(self.rgame.install_path))
        self.ui.install_path.setText(
            self.rgame.install_path if self.rgame.install_path else "N/A"
        )

        self.ui.platform.setText(
            self.rgame.igame.platform
            if self.rgame.is_installed and not self.rgame.is_non_asset
            else self.rgame.default_platform
        )

        self.ui.lbl_grade.setDisabled(
            self.rgame.is_unreal or platform.system() == "Windows"
        )
        self.ui.grade.setDisabled(
            self.rgame.is_unreal or platform.system() == "Windows"
        )
        self.ui.grade.setText(self.steam_grade_ratings[self.rgame.steam_grade()])

        self.ui.install_button.setEnabled(
            (not self.rgame.is_installed or self.rgame.is_non_asset) and self.rgame.is_idle
        )

        self.ui.import_button.setEnabled(
            (not self.rgame.is_installed or self.rgame.is_non_asset) and self.rgame.is_idle
        )

        self.ui.modify_button.setEnabled(
            self.rgame.is_installed
            and (not self.rgame.is_non_asset)
            and self.rgame.is_idle
            and self.rgame.sdl_name is not None
        )

        self.ui.verify_button.setEnabled(
            self.rgame.is_installed and (not self.rgame.is_non_asset) and self.rgame.is_idle
        )
        self.ui.verify_progress.setValue(self.rgame.progress if self.rgame.state == RareGame.State.VERIFYING else 0)
        if self.rgame.state == RareGame.State.VERIFYING:
            self.ui.verify_stack.setCurrentWidget(self.ui.verify_progress_page)
        else:
            self.ui.verify_stack.setCurrentWidget(self.ui.verify_button_page)

        self.ui.repair_button.setEnabled(
            self.rgame.is_installed and (not self.rgame.is_non_asset) and self.rgame.is_idle
            and self.rgame.needs_repair
            and not self.args.offline
        )

        self.ui.move_button.setEnabled(
            self.rgame.is_installed and (not self.rgame.is_non_asset) and self.rgame.is_idle
        )
        self.ui.move_progress.setValue(self.rgame.progress if self.rgame.state == RareGame.State.MOVING else 0)
        if self.rgame.state == RareGame.State.MOVING:
            self.ui.move_stack.setCurrentWidget(self.ui.move_progress_page)
        else:
            self.ui.move_stack.setCurrentWidget(self.ui.move_button_page)

        self.ui.uninstall_button.setEnabled(
            self.rgame.is_installed and (not self.rgame.is_non_asset) and self.rgame.is_idle
        )

        if self.rgame.is_installed and not self.rgame.is_non_asset:
            self.ui.game_actions_stack.setCurrentWidget(self.ui.installed_page)
        else:
            self.ui.game_actions_stack.setCurrentWidget(self.ui.uninstalled_page)

    @pyqtSlot(RareGame)
    def update_game(self, rgame: RareGame):
        if self.rgame is not None:
            if (worker := self.rgame.worker()) is not None:
                if isinstance(worker, VerifyWorker):
                    try:
                        worker.signals.progress.disconnect(self.__on_verify_progress)
                    except TypeError as e:
                        logger.warning(f"{self.rgame.app_name} verify worker: {e}")
                if isinstance(worker, MoveWorker):
                    try:
                        worker.signals.progress.disconnect(self.__on_move_progress)
                    except TypeError as e:
                        logger.warning(f"{self.rgame.app_name} move worker: {e}")
            self.rgame.signals.widget.update.disconnect(self.__update_widget)

        self.rgame = None

        rgame.signals.widget.update.connect(self.__update_widget)
        if (worker := rgame.worker()) is not None:
            if isinstance(worker, VerifyWorker):
                worker.signals.progress.connect(self.__on_verify_progress)
            if isinstance(worker, MoveWorker):
                worker.signals.progress.connect(self.__on_move_progress)

        self.set_title.emit(rgame.app_title)
        self.ui.app_name.setText(rgame.app_name)
        self.ui.dev.setText(rgame.developer)

        if rgame.is_non_asset:
            self.ui.install_button.setText(self.tr("Link/Launch"))
            self.ui.game_actions_stack.setCurrentWidget(self.ui.uninstalled_page)
        else:
            self.ui.install_button.setText(self.tr("Install Game"))

        self.move_game_pop_up.update_game(rgame)

        self.rgame = rgame
        self.__update_widget()



import difflib
import orjson
import os
from datetime import date

import requests

from rare.lgndr.core import LegendaryCore
from rare.utils.paths import data_dir, cache_dir

replace_chars = ",;.:-_ "
url = "https://api.steampowered.com/ISteamApps/GetAppList/v2/"


__steam_ids_json = None
__grades_json = None


def get_rating(core: LegendaryCore, app_name: str):
    global __grades_json
    if __grades_json is None:
        if os.path.exists(p := os.path.join(data_dir(), "steam_ids.json")):
            grades = orjson.loads(open(p).read())
            __grades_json = grades
        else:
            grades = {}
            __grades_json = grades
    else:
        grades = __grades_json

    if not grades.get(app_name):
        game = core.get_game(app_name)
        try:
            steam_id = get_steam_id(game.app_title)
            grade = get_grade(steam_id)
        except:
            return "fail"
        grades[app_name] = {"steam_id": steam_id, "grade": grade}
        with open(os.path.join(data_dir(), "steam_ids.json"), "w") as f:
            f.write(orjson.dumps(grades).decode("utf-8"))
        return grade
    else:
        return grades[app_name].get("grade")


# you should iniciate the module with the game's steam code
def get_grade(steam_code):
    if steam_code == 0:
        return "fail"
    steam_code = str(steam_code)
    url = "https://www.protondb.com/api/v1/reports/summaries/"
    res = requests.get(f"{url}{steam_code}.json")
    try:
        lista = orjson.loads(res.text)
    except orjson.JSONDecodeError:
        return "fail"

    return lista["tier"]


def load_json() -> dict:
    file = os.path.join(cache_dir(), "game_list.json")
    if not os.path.exists(file):
        response = requests.get(url)
        steam_ids = orjson.loads(response.text)["applist"]["apps"]
        ids = {}
        for game in steam_ids:
            ids[game["name"]] = game["appid"]

        with open(file, "w") as f:
            f.write(orjson.dumps(ids).decode("utf-8"))
        return ids
    else:
        return orjson.loads(open(file, "r").read())


def get_steam_id(title: str):
    file = os.path.join(cache_dir(), "game_list.json")
    # workarounds for satisfactory
    title = title.replace("Early Access", "").replace("Experimental", "").strip()
    global __steam_ids_json
    if __steam_ids_json is None:
        if not os.path.exists(file):
            response = requests.get(url)
            ids = {}
            steam_ids = orjson.loads(response.text)["applist"]["apps"]
            for game in steam_ids:
                ids[game["name"]] = game["appid"]
            __steam_ids_json = ids

            with open(file, "w") as f:
                f.write(orjson.dumps(ids).decode("utf-8"))
        else:
            ids = orjson.loads(open(file, "r").read())
            __steam_ids_json = ids
    else:
        ids = __steam_ids_json

    if title in ids.keys():
        steam_name = [title]

    else:
        steam_name = difflib.get_close_matches(title, ids.keys(), n=1)
    if steam_name:
        return ids[steam_name[0]]
    else:
        return 0


def check_time():  # this function check if it's time to update
    file = os.path.join(cache_dir(), "game_list.json")
    json_table = orjson.loads(open(file, "r").read())

    today = date.today()
    day = 0  # it controls how many days it's necessary for an update
    for i in "ymd":
        if i == "d":
            day = 7
        else:
            day = 0
        if int(today.strftime("%" + i)) > int(json_table["data"][i]) + day:
            return 1
        else:
            return 0
